# pre/_multi.py
"""Transformer for states with multiple variables."""

__all__ = [
    "TransformerPipeline",
    "NullTransformer",
    "TransformerMulti",
]

import warnings
import numpy as np

from .. import errors, utils
from ._base import TransformerTemplate, requires_trained


# Horizontal joining ==========================================================
class TransformerPipeline(TransformerTemplate):
    r"""Chain multiple transformers.

    Given :math:`\tau\in\NN` transformers
    :math:`\mathcal{T}_1,\ldots,\mathcal{T}_{\tau}`, this class defines the
    compositional transformer
    :math:`\mathcal{T} = \mathcal{T}_{\tau}\circ\cdots\circ\mathcal{T}_1`.

    Parameters
    ----------
    transformers : tuple of instantiated Transformer objects
        Transformers to be chained together;
        `transformers[0]` is applied first, then `transformers[1]`, and so on.
    name : str or None
        Label for the state variable that this transformer acts on.

    Notes
    -----
    This class connects multiple transformers "horizontally"; see
    :class:`TransformerMulti` to connect multiple transformers "vertically",
    i.e., to assign different transformations for different parts of the state.
    """

    def __init__(self, transformers, name=None):
        """Set the transformers."""
        super().__init__(name=name)

        if isinstance(transformers, list):
            transformers = tuple(transformers)
        else:
            raise TypeError("'transformers' should be a list or tuple")

        statedims = set()
        for i, tf in enumerate(transformers):
            if not isinstance(tf, (TransformerTemplate, TransformerMulti)):
                warnings.warn(
                    f"transformers[{i}] does not inherit from "
                    "TransformerTemplate, unexpected behavior may occur",
                    errors.OpInfWarning,
                )
            statedims.add(tf.state_dimension)
        statedims.discard(None)
        if len(statedims) > 1:
            raise ValueError("transformers have inconsistent state_dimension")

        if len(transformers) == 1:
            warnings.warn(
                "only one transformer provided to TransformerPipeline",
                errors.OpInfWarning,
            )
        self.__transformers = tuple(transformers)

    # Properties --------------------------------------------------------------
    @property
    def transformers(self) -> tuple:
        """Transformers being chained together;
        `transformers[0]` is applied first, then `transformers[1]`, and so on.
        """
        return self.__transformers

    @property
    def num_transformers(self) -> int:
        """Number of transformers chained together."""
        return len(self.transformers)

    @property
    def state_dimension(self) -> int:
        r"""Dimension :math:`n` of the state."""
        for tf in self.transformers:
            if (r := tf.state_dimension) is not None:
                return r
        return None

    def __str__(self):
        lines = super().__str__().split("\n  ")
        lines.append(f"num_transformers: {self.num_transformers}")
        lines.append("transformers")
        for tf in self.transformers:
            tfstr = str(tf).split("\n  ")
            tfstr[0] = f"  {tfstr[0]}"
            lines.append("\n      ".join(tfstr))
        return "\n  ".join(lines)

    def __len__(self):
        return self.num_transformers

    # Main routines -----------------------------------------------------------
    def _chain(self, method, states, inplace=False, **kwargs):
        """Apply the specified method for each transformer (forward)."""
        func = getattr(self.transformers[0], method)
        Y = func(states, inplace=inplace, **kwargs)
        for tf in self.transformers[1:]:
            func = getattr(tf, method)
            Y = func(Y, inplace=True, **kwargs)
        return Y

    def fit_transform(self, states, inplace=False):
        return self._chain("fit_transform", states, inplace=inplace)

    def transform(self, states, inplace=False):
        return self._chain("transform", states, inplace=inplace)

    def inverse_transform(self, states_transformed, inplace=False, locs=None):
        Y = self.transformers[-1].inverse_transform(
            states_transformed,
            inplace=inplace,
            locs=locs,
        )
        for tf in reversed(self.transformers[:-1]):
            Y = tf.inverse_transform(Y, inplace=True, locs=locs)
        return Y

    def transform_ddts(self, ddts, inplace=False):
        return self._chain("transform_ddts", ddts, inplace=inplace)

    # Model persistence -------------------------------------------------------
    def save(self, savefile: str, overwrite: bool = False):
        with utils.hdf5_savehandle(savefile, overwrite) as hf:
            meta = hf.create_dataset("meta", shape=(0,))
            meta.attrs["name"] = str(self.name)
            meta.attrs["num_transformers"] = self.num_transformers
            for i, tf in enumerate(self.transformers):
                tf.save(hf.create_group(f"transformer_{i:0>2d}"))

    @classmethod
    def load(cls, loadfile: str, TransformerClasses):
        """Load a previously saved transformer from an HDF5 file.

        Parameters
        ----------
        loadfile : str
            File where the transformer was stored via :meth:`save()`.
        TransformerClasses : Iterable[type]
            Classes of the transformers for each state variable.

        Returns
        -------
        TransformerPipeline
        """
        with utils.hdf5_loadhandle(loadfile) as hf:
            meta = hf["meta"]
            name = str(meta.attrs["name"])
            num_transformers = int(meta.attrs["num_transformers"])
            if (nclasses := len(TransformerClasses)) != num_transformers:
                raise ValueError(
                    f"file contains {num_transformers:d} transformers "
                    f"but {nclasses:d} classes provided"
                )

            transformers = [
                TransformerClasses[i].load(hf[f"transformer_{i:0>2d}"])
                for i in range(num_transformers)
            ]

            return cls(transformers, name=name)


# Vertical joining ============================================================
class NullTransformer(TransformerTemplate):
    r"""Identity transformation :math:`\q\mapsto\q`.

    This transformer can be used in conjunction with :class:`TransformerMulti`
    if separate transformations are desired for individual state variables but
    one of those state variable is to remain unchanged.

    Parameters
    ----------
    name : str or None
        Label for the state variable that this transformer "acts" on.
    """

    # Main routines -----------------------------------------------------------
    def fit(self, states):
        """Set the state dimension.

        Parameters
        ----------
        states : (n, k) ndarray
            Matrix of `k` `n`-dimensional snapshots.

        Returns
        -------
        self
        """
        self.state_dimension = states.shape[0]
        return self

    def fit_transform(self, states, inplace: bool = True):
        """Do nothing but set the state dimension; this transformation does not
        affect the states.

        Parameters
        ----------
        states : (n, ...) ndarray
            Matrix of `n`-dimensional snapshots, or a single snapshot.
        inplace : bool
            If ``True`` (default), return ``states``.
            If ``False``, return a copy of ``states``.

        Returns
        -------
        states: (n, ...) ndarray
            State snapshots, or a copy of them if ``inplace=False``.
        """
        self.fit(states)
        return states if inplace else states.copy()

    def transform(self, states, inplace: bool = False):
        """Do nothing; this transformation does not affect the states.

        Parameters
        ----------
        states : (n, ...) ndarray
            Matrix of `n`-dimensional snapshots, or a single snapshot.
        inplace : bool
            If ``True`` (default), return ``states``.
            If ``False``, return a copy of ``states``.

        Returns
        -------
        states: (n, ...) ndarray
            State snapshots, or a copy of them if ``inplace=False``.

        Raises
        ------
        ValueError
            If the ``states`` do not align with the :attr:`state_dimension`.
        """
        self._check_shape(states)
        return states if inplace else states.copy()

    def transform_ddts(self, ddts, inplace: bool = True):
        r"""Do nothing; this transformation does not affect derivatives.

        Parameters
        ----------
        ddts : (n, ...) ndarray
            Matrix of `n`-dimensional snapshot time derivatives, or a
            single snapshot time derivative.
        inplace : bool
            If ``True`` (default), return ``ddts``.
            If ``False``, return a create a copy of ``ddts``.

        Returns
        -------
        ddts : (n, ...) ndarray
            Snapshot time derivatives, or a copy of them if ``inplace=False``.

        Raises
        ------
        ValueError
            If the ``states`` do not align with the :attr:`state_dimension`.
        """
        self._check_shape(ddts)
        return ddts if inplace else ddts.copy()

    def inverse_transform(self, states_transformed, inplace=False, locs=None):
        """Do nothing; this transformation does not affect the states.

        Parameters
        ----------
        states_transformed : (n, ...) or (p, ...)  ndarray
            Matrix of `n`-dimensional transformed snapshots, or a single
            transformed snapshot.
        inplace : bool
            If ``True`` (default), return ``states_transformed``.
            If ``False``, return a create a copy of ``states_transformed``.
        locs : slice or (p,) ndarray of integers or None
            If given, assume ``states_transformed`` contains the transformed
            snapshots at only the `p` indices described by ``locs``.

        Returns
        -------
        states_transformed: (n, ...) or (p, ...) ndarray
            Transformed states, or a copy of them if ``inplace=False``.

        Raises
        ------
        ValueError
            If the ``states_transformed`` do not align with the ``locs`` (when
            provided) or the :attr:`state_dimension` (when ``locs`` is not
            provided).
        """
        if locs is not None:
            locs = self._check_locs(locs, states_transformed)
        else:
            self._check_shape(states_transformed)
        return states_transformed if inplace else states_transformed.copy()

    # Model persistence -------------------------------------------------------
    def save(self, savefile, overwrite=False):
        with utils.hdf5_savehandle(savefile, overwrite) as hf:
            meta = hf.create_dataset("meta", shape=(0,))
            meta.attrs["name"] = str(self.name)
            if (n := self.state_dimension) is not None:
                meta.attrs["state_dimension"] = n

    @classmethod
    def load(cls, loadfile):
        with utils.hdf5_loadhandle(loadfile) as hf:
            meta = hf["meta"]
            name = meta.attrs["name"]
            ntf = cls(name=(None if name == "None" else name))
            if "state_dimension" in meta.attrs:
                ntf.state_dimension = int(meta.attrs["state_dimension"])
            return ntf


class TransformerMulti:
    r"""Join transformers together for states with multiple variables.

    This class is for states that can be written (after discretization) as

    .. math::
       \q = \left[\begin{array}{c}
       \q_{0} \\ \q_{1} \\ \vdots \\ \q_{n_q - 1}
       \end{array}\right]
       \in \RR^{n},

    where each :math:`\q_{i} \in \RR^{n_i}` represents a single discretized
    state variable of dimension :math:`n_i \in \NN`. The dimension of the
    joint state :math:`\q` is :math:`n = \sum_{i=0}^{n_q - 1} n_i`, i.e.,
    ``state_dimension = sum(variable_sizes)``.
    Individual transformers are calibrated for each state variable.

    Parameters
    ----------
    transformers : list
        Initialized (but not necessarily trained) transformer objects,
        one for each state variable. Entries of this list can be ``None``,
        in which case a :class:`NullTransformer` is used.
    variable_sizes : list or None
        Dimensions for each state variable, :math:`n_0,\ldots,n_{n_q-1}`.
        If ``None`` (default), set :math:`n_i` to
        ``transformers[i].state_dimension``; if any of these are ``None``,
        assume all state variables have the same dimension, i.e.,
        :math:`n_0 = n_1 = \cdots = n_x \in \NN` with :math:`n_x` to be
        determined in :meth:`fit`. In this case, :math:`n = n_q n_x`.

    Notes
    -----
    This class connects multiple transformers "vertically" by assigning
    different transformations for different parts of the state; see
    :class:`TransformerPipeline` to connect multiple transformers
    "horizontally", i.e., to perform one transformation after another on the
    entire state.
    """

    def __init__(self, transformers, variable_sizes=None):
        """Initialize the transformers."""
        if (num_variables := len(transformers)) == 0:
            raise ValueError("at least one transformer required")
        elif num_variables == 1:
            warnings.warn("only one variable detected", errors.OpInfWarning)

        # Check inheritance and set default variable names.
        tfs = [NullTransformer() if tf is None else tf for tf in transformers]
        for i, tf in enumerate(tfs):
            if not isinstance(tf, TransformerTemplate):
                warnings.warn(
                    f"transformers[{i}] does not inherit from "
                    "TransformerTemplate, unexpected behavior may occur",
                    errors.OpInfWarning,
                )
            if tf.name is None:
                tf.name = f"variable {i}"

        # Check variable sizes.
        if variable_sizes is not None:
            if len(variable_sizes) != num_variables:
                raise ValueError("len(variable_sizes) != len(transformers)")
            for tf, ni in zip(tfs, variable_sizes):
                if tf.state_dimension is None:
                    tf.state_dimension = ni
                elif (tf_n := tf.state_dimension) != ni:
                    raise ValueError(
                        f"transformers[{i}].state_dimension = {tf_n} "
                        f"!= {ni} = variable_sizes[{i}]"
                    )

        # Store transformers and set slice dimensions.
        self.__nq = num_variables
        self.__transformers = tuple(tfs)

    # Properties --------------------------------------------------------------
    @property
    def transformers(self) -> tuple:
        """Transformers for each state variable."""
        return self.__transformers

    @property
    def num_variables(self) -> int:
        r"""Number of state variables :math:`n_q \in \NN`."""
        return self.__nq

    @property
    def variable_names(self) -> tuple:
        """Names for each state variable."""
        return tuple(tf.name for tf in self.transformers)

    @property
    def variable_sizes(self) -> tuple:
        r"""Dimensions of each state variable,
        :math:`n_0,\ldots,n_{n_q-1}\in\NN`.
        """
        return tuple(tf.state_dimension for tf in self.transformers)

    @property
    def state_dimension(self) -> int:
        r"""Total dimension :math:`n = \sum_{i=0}^{n_q-1} n_i \in \NN` of the
        joint state.
        """
        if None in (sizes := self.variable_sizes):
            return None
        return sum(sizes)

    # Magic methods -----------------------------------------------------------
    def __len__(self) -> int:
        """Length = number of state variables."""
        return self.num_variables

    def __getitem__(self, key):
        """Get the transformer for variable i."""
        if key in self.variable_names:
            key = self.variable_names.index(key)
        return self.transformers[key]

    def __eq__(self, other) -> bool:
        """Test two TransformerMulti objects for equality."""
        if not (
            isinstance(other, self.__class__)
            or isinstance(self, other.__class__)
        ):
            return False
        if self.num_variables != other.num_variables:
            return False
        return all(
            t1 == t2 for t1, t2 in zip(self.transformers, other.transformers)
        )

    def __str__(self) -> str:
        """String representation: str() of each transformer."""
        lines = [self.__class__.__name__]
        if (n := self.state_dimension) is not None:
            lines.append(f"state_dimension: {n}")
        lines.append(f"num_variables:   {self.num_variables}")
        lines.append("transformers")
        for tf in self.transformers:
            tfstr = str(tf).split("\n  ")
            tfstr[0] = f"  {tfstr[0]}"
            lines.append("\n      ".join(tfstr))
        return "\n  ".join(lines)

    def __repr__(self) -> str:
        """Unique ID + string representation."""
        return utils.str2repr(self)

    # Access convenience methods ----------------------------------------------
    @requires_trained
    def _check_shape(self, states):
        """Verify the shape of the snapshot set Q."""
        if (nQ := len(states)) != (n := self.state_dimension):
            raise errors.DimensionalityError(
                f"len(states) = {nQ:d} != {n:d} = state_dimension"
            )

    def _slices(self, varindex=None):
        """Get slices for one or all state variable(s).

        Parameters
        ----------
        varindex : int
            Index of the variable to get a slice for.
            If ``None`` (default), get slices for all state variables.

        Returns
        -------
        slice or tuple
            Slice for the state variable at index ``varindex``
            or a tuple of slices for all state variables.
        """
        dimsum = np.cumsum((0,) + self.variable_sizes)
        if varindex is not None:
            return slice(dimsum[varindex], dimsum[varindex + 1])

        return tuple(
            [
                slice(dimsum[i], dimsum[i + 1])
                for i in range(self.num_variables)
            ]
        )

    def get_var(self, var, states):
        """Extract a single variable from the joint state.

        Parameters
        ----------
        var : int or str
            Index or name of the variable to extract.
        states : (n, ...) ndarray
            Joint state vector or snapshot matrix.

        Returns
        -------
        state_variable : (n_i, ...) ndarray
            One state variable, extracted from ``states``.
        """
        self._check_shape(states)
        if var in (names := self.variable_names):
            var = names.index(var)
        return states[self._slices(var)]

    def split(self, states):
        """Split the joint state into the individual state variables.

        Parameters
        ----------
        states : (n, ...) ndarray
            Joint state vector or snapshot matrix.

        Returns
        -------
        state_variable : list of nq (nx, ...) ndarrays
            Individual state variables, extracted from ``states``.
        """
        self._check_shape(states)
        return [states[s] for s in self._slices()]

    # Main routines -----------------------------------------------------------
    def fit(self, states):
        """Learn (but do not apply) the transformation.

        Parameters
        ----------
        states : (n, k) ndarray
            Matrix of `k` `n`-dimensional snapshots.

        Returns
        -------
        self
        """
        self.fit_transform(states, inplace=False)
        return self

    def fit_transform(self, states, inplace=False):
        """Learn and apply the transformation.

        Parameters
        ----------
        states : (n, k) ndarray
            Matrix of `k` `n`-dimensional snapshots.
            The first ``variable_sizes[0]`` entries correspond to the first
            state variable, the next ``variable_sizes[1]`` entries correspond
            to the second state variable, and so on.
            If ``variable_sizes`` are not yet prescribed, assume that each
            state variable has the same dimension.
        inplace : bool
            If ``True``, overwrite ``states`` data during transformation.
            If ``False``, create a copy of the data to transform.

        Returns
        -------
        states_transformed: (n, k) ndarray
            Matrix of `k` transformed `n`-dimensional snapshots.
        """
        if self.state_dimension is None:
            # Assume all state dimensions are equal.
            nx, remainder = divmod(states.shape[0], self.num_variables)
            if remainder != 0:
                raise errors.DimensionalityError(
                    "len(states) must be evenly divisible by "
                    f"the number of variables n_q = {self.num_variables}"
                )
            for tf in self.transformers:
                TransformerTemplate.state_dimension.fset(tf, nx)

        new_states = np.concatenate(
            [
                transformer.fit_transform(state_variable, inplace=inplace)
                for transformer, state_variable in zip(
                    self.transformers, self.split(states)
                )
            ],
            axis=0,
        )
        return states if inplace else new_states

    def transform(self, states, inplace=False):
        """Apply the learned transformation.

        Parameters
        ----------
        states : (n, ...) ndarray
            Matrix of `n`-dimensional snapshots, or a single snapshot.
        inplace : bool
            If ``True``, overwrite ``states`` during transformation.
            If ``False``, create a copy of the data to transform.

        Returns
        -------
        states_transformed: (n, ...) ndarray
            Matrix of `n`-dimensional transformed snapshots, or a single
            transformed snapshot.
        """
        states_transformed = np.concatenate(
            [
                tf.transform(Q, inplace=inplace)
                for tf, Q in zip(self.transformers, self.split(states))
            ]
        )
        return states if inplace else states_transformed

    def transform_ddts(self, ddts, inplace: bool = False):
        r"""Apply the learned transformation to snapshot time derivatives.

        If the transformation is denoted by :math:`\mathcal{T}(q)`,
        this function implements :math:`\mathcal{T}'` such that
        :math:`\mathcal{T}'(\ddt q) = \ddt \mathcal{T}(q)`.

        Parameters
        ----------
        ddts : (n, ...) ndarray
            Matrix of `n`-dimensional snapshot time derivatives, or a
            single snapshot time derivative.
        inplace : bool
            If True, overwrite ``ddts`` during the transformation.
            If False, create a copy of the data to transform.

        Returns
        -------
        ddts_transformed : (n, ...) ndarray
            Transformed snapshot time derivatives.
        """
        ddts_transformed = [
            tf.transform_ddts(dQdt, inplace=inplace)
            for tf, dQdt in zip(self.transformers, self.split(ddts))
        ]
        if any(dt is NotImplemented for dt in ddts_transformed):
            return NotImplemented
        return ddts if inplace else np.concatenate(ddts_transformed)

    @requires_trained
    def inverse_transform(self, states_transformed, inplace=False, locs=None):
        """Apply the inverse of the learned transformation.

        Parameters
        ----------
        states_transformed : (n, ...) or (num_variables*p, ...) ndarray
            Matrix of `n`-dimensional transformed snapshots, or a single
            transformed snapshot.
        inplace : bool
            If ``True``, overwrite ``states_transformed`` during the inverse
            transformation. If ``False``, create a copy of the data to
            untransform.
        locs : slice or (p,) ndarray of integers or None
            If given, assume ``states_transformed`` contains the `p` entries
            of each transformed state variable at the indices specified by
            ``locs``. This option requires each state variable to have the
            same dimension.

        Returns
        -------
        states_untransformed: (n, ...) or (num_variables*p, ...) ndarray
            Matrix of `n`-dimensional untransformed snapshots, or the
            :math:`n_q p` entries of such at the indices specified by ``locs``.
        """
        if locs is not None:
            if len(set(self.variable_sizes)) > 1:
                raise ValueError(
                    "'locs != None' requires that "
                    "all transformers have the same state_dimension"
                )
            variables_transformed = np.split(
                states_transformed,
                self.num_variables,
                axis=0,
            )
        else:
            variables_transformed = self.split(states_transformed)

        states = np.concatenate(
            [
                tf.inverse_transform(Q, inplace=inplace, locs=locs)
                for tf, Q in zip(self.transformers, variables_transformed)
            ]
        )
        return states_transformed if inplace else states

    # Model persistence -------------------------------------------------------
    def save(self, savefile: str, overwrite: bool = False):
        with utils.hdf5_savehandle(savefile, overwrite) as hf:
            hf.create_dataset("num_variables", data=[self.num_variables])
            for i, tf in enumerate(self.transformers):
                tf.save(hf.create_group(f"variable{i}"))

    @classmethod
    def load(cls, loadfile: str, TransformerClasses):
        """Load a previously saved transformer from an HDF5 file.

        Parameters
        ----------
        loadfile : str
            File where the transformer was stored via :meth:`save()`.
        TransformerClasses : Iterable[type]
            Classes of the transformers for each state variable.

        Returns
        -------
        TransformerMulti
        """
        with utils.hdf5_loadhandle(loadfile) as hf:
            num_variables = int(hf["num_variables"][0])

            if isinstance(TransformerClasses, type):
                TransformerClasses = [TransformerClasses] * num_variables
            if (nclasses := len(TransformerClasses)) != num_variables:
                raise ValueError(
                    f"file contains {num_variables:d} transformers "
                    f"but {nclasses:d} classes provided"
                )

            transformers = [
                TransformerClasses[i].load(hf[f"variable{i}"])
                for i in range(num_variables)
            ]

            return cls(transformers)

    # Verification ------------------------------------------------------------
    def verify(self, tol: float = 1e-4):
        r"""Verify that :meth:`transform()` and :meth:`inverse_transform()`
        are consistent and that :meth:`transform_ddts()`, if implemented in
        each transformer, is consistent with :meth:`transform()`.

        * The :meth:`transform()` / :meth:`inverse_transform()` consistency
          check verifies that
          ``inverse_transform(transform(states)) == states``.
        * The :meth:`transform_ddts()` consistency check uses
          :meth:`opinf.ddt.ddt()` to estimate the time derivatives of the
          states and the transformed states, then verfies that the relative
          difference between
          ``transform_ddts(opinf.ddt.ddt(states, t))`` and
          ``opinf.ddt.ddt(transform(states), t)`` is less than ``tol``.
          If this check fails, consider using a finer time mesh.

        Parameters
        ----------
        tol : float > 0
            Tolerance for the finite difference check of
            :meth:`transform_ddts()`.
            Only used if :meth:`transform_ddts()` is implemented.
        """
        for tf, name in zip(self.transformers, self.variable_names):
            print(f"{name}:", end="\t")
            tf.verify(tol)
        return TransformerTemplate.verify(self, tol)

    def _verify_locs(self, states, states_transformed):
        """Verify :meth:`inverse_transform()` with ``locs != None``."""
        if len(sizes := set(self.variable_sizes)) != 1:
            return  # Cannot use locs unless all variable sizes equal.
        nx = sizes.pop()
        locs = np.sort(np.random.choice(nx, size=(nx // 3), replace=False))

        states_at_locs = np.concatenate([Q[locs] for Q in self.split(states)])
        states_transformed_at_locs = np.concatenate(
            [Qt[locs] for Qt in self.split(states_transformed)]
        )
        states_recovered_at_locs = self.inverse_transform(
            states_transformed_at_locs,
            locs=locs,
        )

        if states_recovered_at_locs.shape != states_at_locs.shape:
            raise errors.VerificationError(  # pragma: no cover
                "inverse_transform(states_transformed_at_locs, locs).shape "
                "!= states_at_locs.shape"
            )
        if not np.allclose(states_recovered_at_locs, states_at_locs):
            raise errors.VerificationError(  # pragma: no cover
                "transform() and inverse_transform() are not inverses "
                "(locs != None)"
            )



# utils/_mpl_config.py
"""Matplotlib configuration for documenation notebooks."""

__all__ = [
    "mpl_config",
]

import matplotlib.pyplot as plt

from ._timer import TimedBlock


def mpl_config():
    """Set matplotlib and pandas configuration defaults for the
    documentation notebooks.
    """
    # Matplotlib customizations.
    plt.rc("axes.spines", right=False, top=False)
    plt.rc("figure", dpi=300, figsize=(9, 3))
    plt.rc("font", family="serif")
    plt.rc("legend", edgecolor="none", frameon=False)
    plt.rc("text", usetex=True)

    # Turn off stdout flushing to keep notebook outputs together.
    TimedBlock.rebuffer = True

    # Pandas display options.
    try:
        import pandas as pd
    except ModuleNotFoundError as ex:  # pragma: no cover
        if ex.args[0] != "No module named 'pandas'":
            raise
    else:  # pragma: no cover
        pd.options.display.float_format = "{:.4%}".format



# lstsq/_total.py
"""Operator Inference total least-squares solver."""

__all__ = [
    "TotalLeastSquaresSolver",
]

import types
import numpy as np
import scipy.linalg as la

from .. import utils
from ._base import SolverTemplate, _require_trained


class TotalLeastSquaresSolver(SolverTemplate):
    r"""Solve the total least-squares problem without any
    regularization, i.e.,

    .. math::
       \argmin_{\Ohat}
       \left\|[~\Delta_\D~~\Delta_{\Z\trp}~]\right\|_{F}^{2}
       \quad\text{such that}\quad
       (\D + \Delta_\D)\Ohat\trp = \Z\trp + \Delta_{\Z\trp}.

    The solution is computed via the singular value decomposition of the
    augmented :math:`k \times (d + r)` matrix :math:`[~\D~~\Z\trp~]`: if
    :math:`\bfPhi\bfSigma\bfPsi\trp = [~\D~~\Z\trp~]`, write the
    :math:`(d + r) \times (d + r)` right singular vector matrix
    :math:`\bfPsi` as

    .. math::
       \bfPsi = \left[\begin{array}{cc}
           \bfPsi_{\D\D} & \bfPsi_{\D\Z} \\
           \bfPsi_{\Z\D} & \bfPsi_{\Z\Z}
       \end{array}\right],

    where
    :math:`\bfPsi_{\D\D}` is :math:`d \times d`,
    :math:`\bfPsi_{\D\Z}` is :math:`d \times r`,
    :math:`\bfPsi_{\Z\D}` is :math:`r \times d`, and
    :math:`\bfPsi_{\Z\Z}` is :math:`r \times r`.
    Then the optimal solution is given by
    :math:`\Ohat = -\bfPsi_{\Z\Z}^{-\mathsf{T}}\bfPsi_{\D\Z}\trp`.

    Parameters
    ----------
    lapack_driver : str
        LAPACK routine for computing the SVD. See :func:`scipy.linalg.svd()`.
    """

    def __init__(self, lapack_driver: str = "gesdd"):
        SolverTemplate.__init__(self)
        self.__options = types.MappingProxyType(
            dict(full_matrices=False, lapack_driver=lapack_driver)
        )

    # Properties --------------------------------------------------------------
    @property
    def options(self) -> dict:
        """Keyword arguments for :func:`scipy.linalg.svd()`."""
        return self.__options

    def __str__(self) -> str:
        """String representation: dimensions + solver options."""
        start = SolverTemplate.__str__(self)
        if self.data_matrix is not None:
            lines = start.split("\n")
            lines.insert(
                4,
                f"    Augmented condition number: {self._augcond:.4e}",
            )
            start = "\n".join(lines)
        kwargs = self._print_kwargs(self.options)
        return start + f"\n  SVD solver: scipy.linalg.svd({kwargs})"

    # Main methods ------------------------------------------------------------
    def fit(self, data_matrix: np.ndarray, lhs_matrix: np.ndarray):
        r"""Verify dimensions, compute the singular value decomposition of
        the data matrix, and solve the problem.

        Parameters
        ----------
        data_matrix : (k, d) ndarray
            Data matrix :math:`\D`.
        lhs_matrix : (r, k) or (k,) ndarray
            "Left-hand side" data matrix :math:`\Z` (not its transpose!).
            If one-dimensional, assume :math:`r = 1`.
        """
        SolverTemplate.fit(self, data_matrix, lhs_matrix)
        if self.k < self.d + self.r:
            raise ValueError(
                "total least-squares system is underdetermined, k > d + r "
                f"is required (k = {self.k}, d = {self.d}, r = {self.r})"
            )

        # Compute the SVD of the concatenation [D Z^T].
        D_Zt = np.hstack((self.data_matrix, self.lhs_matrix.T))
        Phi, _svals, PsiT = la.svd(D_Zt, **self.options)
        self._augcond = _svals.max() / _svals.min()

        # Extract the relevant blocks of the singular vector matrices.
        Psi_ZZt = PsiT[self.d :, self.d :]
        Psi_DZt = PsiT[self.d :, : self.d]
        Phi_Zt_Sig_ZT = Phi[:, self.d :] * _svals[self.d :]

        # Solve the problem and compute the residuals.
        self._Ohat = -la.solve(Psi_ZZt, Psi_DZt)
        Deltas = -Phi_Zt_Sig_ZT @ np.hstack((Psi_ZZt, Psi_DZt))
        self._norm_of_errors = la.norm(Deltas)

        return self

    @_require_trained
    def solve(self) -> np.ndarray:
        r"""Return the total least-squares solution to the Operator Inference
        regression.

        Returns
        -------
        Ohat : (r, d) ndarray
            Operator matrix :math:`\Ohat` (not its transpose!).
        """
        return self._Ohat

    # Post-processing ---------------------------------------------------------
    @property
    @_require_trained
    def augcond(self) -> float:
        r""":math:`2`-norm condition number of the augmented data matrix
        :math:`[~\D~~\Z\trp~]`.
        """
        return self._augcond

    @property
    @_require_trained
    def error(self) -> float:
        r"""Frobenius norm of the error matrices,
        :math:`\|[~\Delta_\D~~\Delta_{\Z\trp}~\|_F`.
        """
        return self._norm_of_errors

    # Persistence -------------------------------------------------------------
    def save(self, savefile: str, overwrite: bool = False):
        """Serialize the solver, saving it in HDF5 format.
        The model can be recovered with the :meth:`load()` class method.

        Parameters
        ----------
        savefile : str
            File to save to, with extension ``.h5`` (HDF5).
        overwrite : bool
            If ``True`` and the specified ``savefile`` already exists,
            overwrite the file. If ``False`` (default) and the specified
            ``savefile`` already exists, raise an error.
        """
        with utils.hdf5_savehandle(savefile, overwrite) as hf:
            self._save_dict(hf, "options")
            if self.data_matrix is not None:
                for attr in ("data_matrix", "lhs_matrix", "_Ohat"):
                    hf.create_dataset(attr, data=getattr(self, attr))
                for attr in ("_augcond", "_norm_of_errors"):
                    hf.create_dataset(attr, data=[getattr(self, attr)])

    @classmethod
    def load(cls, loadfile: str):
        """Load a serialized solver from an HDF5 file, created previously from
        the :meth:`save()` method.

        Parameters
        ----------
        loadfile : str
            Path to the file where the model was stored via :meth:`save()`.

        Returns
        -------
        solver : L2Solver
            Loaded solver.
        """
        with utils.hdf5_loadhandle(loadfile) as hf:
            options = cls._load_dict(hf, "options")
            solver = cls(lapack_driver=options["lapack_driver"])
            if "data_matrix" in hf:
                D = hf["data_matrix"][:]
                Z = hf["lhs_matrix"][:]
                SolverTemplate.fit(solver, D, Z)
                solver._Ohat = hf["_Ohat"][:]
                for attr in ("_augcond", "_norm_of_errors"):
                    setattr(solver, attr, float(hf[attr][0]))
        return solver

    def copy(self):
        """Make a copy of the solver."""
        solver = self.__class__(lapack_driver=self.options["lapack_driver"])
        if self.data_matrix is not None:
            SolverTemplate.fit(solver, self.data_matrix, self.lhs_matrix)
            for attr in ("_Ohat", "_augcond", "_norm_of_errors"):
                setattr(solver, attr, getattr(self, attr))
        return solver



# roms/_base.py
"""Base for ROM classes."""

__all__ = []

import abc
import warnings
import numpy as np

from .. import errors, post, utils
from .. import lift, pre, basis as _basis, ddt
from ..models import _utils as modutils


def _identity(x):
    return x


class _BaseROM(abc.ABC):
    r"""Base class for reduced-order models.

    ROM classes connect classes from the various submodules to form a complete
    reduced-order modeling workflow:

    High-dimensional data
    :math:`\to` transformed / preprocessed data
    :math:`\to` compressed data
    :math:`\to` low-dimensional model.

    Parameters
    ----------
    model : :mod:`opinf.models` object
        System model.
    lifter : :mod:`opinf.lift` object or None
        Lifting transformation.
    transformer : :mod:`opinf.pre` object or None
        Preprocesser.
    basis : :mod:`opinf.basis` object or None
        Dimensionality reducer.
    ddt_estimator : :mod:`opinf.ddt` object or None
        Time derivative estimator.
        Ignored if ``model`` is not time continuous.
    """

    def __init__(self, model, lifter, transformer, basis, ddt_estimator):
        """Store attributes. Child classes should verify model type."""
        self.__model = model

        # Verify lifter.
        if not (lifter is None or isinstance(lifter, lift.LifterTemplate)):
            warnings.warn(
                "lifter not derived from LifterTemplate, "
                "unexpected behavior may occur",
                errors.OpInfWarning,
            )
        self.__lifter = lifter

        # Verify transformer.
        if not (
            transformer is None
            or isinstance(
                transformer,
                (pre.TransformerTemplate, pre.TransformerMulti),
            )
        ):
            warnings.warn(
                "transformer not derived from TransformerTemplate "
                "or TransformerMulti, unexpected behavior may occur",
                errors.OpInfWarning,
            )
        self.__transformer = transformer

        # Verify basis.
        if not (
            basis is None
            or isinstance(basis, (_basis.BasisTemplate, _basis.BasisMulti))
        ):
            warnings.warn(
                "basis not derived from BasisTemplate or BasisMulti, "
                "unexpected behavior may occur",
                errors.OpInfWarning,
            )
        self.__basis = basis

        # Verify ddt_estimator.
        if (ddt_estimator is not None) and not self._iscontinuous:
            warnings.warn(
                "ddt_estimator ignored for discrete models",
                errors.OpInfWarning,
            )
            ddt_estimator = None
        if not (
            ddt_estimator is None
            or isinstance(ddt_estimator, ddt.DerivativeEstimatorTemplate)
        ):
            warnings.warn(
                "ddt_estimator not derived from DerivativeEstimatorTemplate, "
                "unexpected behavior may occur",
                errors.OpInfWarning,
            )
        self.__ddter = ddt_estimator

    # Properties --------------------------------------------------------------
    @property
    def lifter(self):
        """Lifting transformation."""
        return self.__lifter

    @property
    def transformer(self):
        """Preprocesser."""
        return self.__transformer

    @property
    def basis(self):
        """Dimensionality reducer."""
        return self.__basis

    @property
    def ddt_estimator(self):
        """Time derivative estimator."""
        return self.__ddter

    @property
    def model(self):
        """System model."""
        return self.__model

    @property
    def _iscontinuous(self):
        """``True`` if the model is time continuous (semi-discrete),
        ``False`` if the model if fully discrete.
        """
        return modutils.is_continuous(self.model)

    # Printing ----------------------------------------------------------------
    def __str__(self):
        """String representation."""
        lines = []
        for label, obj in [
            ("lifter", self.lifter),
            ("transformer", self.transformer),
            ("basis", self.basis),
            ("ddt_estimator", self.ddt_estimator),
            ("model", self.model),
        ]:
            if obj is not None:
                lines.append(f"{label}: {str(obj)}")

        body = "\n  ".join("\n".join(lines).split("\n"))
        return f"{self.__class__.__name__}\n  {body}"

    def __repr__(self):
        """Repr: address + string representatation."""
        return utils.str2repr(self)

    # Mappings between original and latent state spaces -----------------------
    def encode(
        self,
        states,
        lhs=None,
        inplace: bool = False,
        *,
        fit_transformer: bool = False,
        fit_basis: bool = False,
    ):
        """Map high-dimensional data to its low-dimensional representation.

        Parameters
        ----------
        states : (n,) or (n, k) ndarray
            State snapshots in the original state space.
        lhs : (n,) or (n, k) ndarray or None
            Left-hand side regression data.

            - If the model is time continuous, these are the time derivatives
              of the state snapshots.
            - If the model is fully discrete, these are the "next states"
              corresponding to the state snapshots.
        inplace : bool
            If ``True``, modify the ``states`` and ``lhs`` in-place in the
            preprocessing transformation (if applicable).

        Returns
        -------
        states_encoded : (r,) or (r, k) ndarray
            Low-dimensional representation of ``states``
            in the latent reduced state space.
        lhs_encoded : (r,) or (r, k) ndarray
            Low-dimensional representation of ``lhs``
            in the latent reduced state space.
            **Only returned** if ``lhs`` is not ``None``.
        """
        # Lifting.
        if self.lifter is not None:
            if lhs is not None:
                if self._iscontinuous:
                    lhs = self.lifter.lift_ddts(states, lhs)
                else:
                    lhs = self.lifter.lift(lhs)
            states = self.lifter.lift(states)

        # Preprocessing.
        if self.transformer is not None:
            if fit_transformer:
                states = self.transformer.fit_transform(
                    states,
                    inplace=inplace,
                )
            else:
                states = self.transformer.transform(states, inplace=inplace)
            if lhs is not None:
                if self._iscontinuous:
                    lhs = self.transformer.transform_ddts(lhs, inplace=inplace)
                else:
                    lhs = self.transformer.transform(lhs, inplace=inplace)

        # Dimensionality reduction.
        if self.basis is not None:
            if fit_basis:
                self.basis.fit(states)
            states = self.basis.compress(states)
            if lhs is not None:
                lhs = self.basis.compress(lhs)

        if lhs is not None:
            return states, lhs
        return states

    def decode(self, states_encoded, locs=None):
        """Map low-dimensional data to the original state space.

        Parameters
        ----------
        states_encoded : (r, ...) ndarray
            Low-dimensional state or states
            in the latent reduced state space.
        locs : slice or (p,) ndarray of integers or None
            If given, return the decoded state at only the p specified
            locations (indices) described by ``locs``.

        Returns
        -------
        states_decoded : (n, ...) ndarray
            Version of ``states_compressed`` in the original state space.
        """
        inplace = False

        # Reverse dimensionality reduction.
        states = states_encoded
        if self.basis is not None:
            inplace = True
            states = self.basis.decompress(states, locs=locs)

        # Reverse preprocessing.
        if self.transformer is not None:
            states = self.transformer.inverse_transform(
                states,
                inplace=inplace,
                locs=locs,
            )

        # Reverse lifting.
        if self.lifter is not None:
            states = self.lifter.unlift(states)

        return states

    def project(self, states):
        """Project a high-dimensional state vector to the subset of the
        high-dimensional space that can be represented by the basis.

        This is done by

        1. expressing the state in low-dimensional latent coordinates, then
        2. reconstructing the high-dimensional state corresponding to those
           coordinates.

        In other words, ``project(Q)`` is equivalent to ``decode(encode(Q))``.

        Parameters
        ----------
        states : (n, ...) ndarray
            Matrix of `n`-dimensional state vectors, or a single state vector.

        Returns
        -------
        state_projected : (n, ...) ndarray
            Matrix of `n`-dimensional projected state vectors, or a single
            projected state vector.
        """
        return self.decode(self.encode(states))

    # Training ----------------------------------------------------------------
    def _check_fit_args(self, lhs, inputs):
        """Verify required arguments for :meth:`fit()`."""

        # Make sure lhs is given if required.
        if lhs is None and self._iscontinuous and self.ddt_estimator is None:
            raise ValueError(
                "argument 'lhs' required when model is time-continuous"
                " and ddt_estimator=None"
            )

        # Make sure inputs are passed in correctly when requried.
        if inputs is None and self.model._has_inputs:
            raise ValueError(
                "argument 'inputs' required (model depends on external inputs)"
            )

    @staticmethod
    def _fix_single_trajectory(states, lhs, inputs):
        """If data comes from a single trajectory, ``states``, ``lhs``, and/or
        ``inputs`` may be arrays instead of lists of arrays. This method casts
        each input as a list of arrays if ``states`` is an array.
        """
        if states[0].ndim == 1:
            states = [states]
            if lhs is not None:
                lhs = [lhs]
            if inputs is not None:
                inputs = [inputs]
        return states, lhs, inputs

    def _process_training_data(
        self,
        states,
        lhs,
        inputs,
        fit_transformer: bool,
        fit_basis: bool,
    ):
        """Process data used to train the model by lifting, transforming,
        reducing, and/or estimating time derivatives.

        Child classes should overwrite this method to include a call to
        the ``fit()`` method of :attr:`model`.

        Parameters
        ----------
        states : list of s (n, k_i) ndarrays
            State snapshots in the original state space. Each array
            ``states[i]`` is data for a single trajectory; each column
            ``states[i][:, j]`` is one snapshot.
        lhs : list of s (n, k_i) ndarrays or None
            Left-hand side regression data. Each array ``lhs[i]`` is the data
            corresponding to ``states[i]``; each column ``lhs[i][:, j]``
            corresponds to the snapshot ``states[i][:, j]``.

            - If the model is time continuous, these are the time derivatives
              of the state snapshots.
            - If the model is fully discrete, these are the "next states"
              corresponding to the state snapshots.

            If ``None``, these are estimated using :attr:`ddt_estimator`
            (time continuous) or extracted from ``states`` (fully discrete).
        inputs : list of s (m, k_i) ndarrays or None
            Input training data. Each array ``inputs[i]`` is the data
            corresponding to parameter value ``parameters[i]``; each column
            ``inputs[i][:, j]`` corresponds to the snapshot ``states[:, j]``.
            May be a two-dimensional array if :math:`m=1` (scalar input).
            Only required if one or more model operators depend on inputs.
        fit_transformer : bool
            If ``True``, calibrate the preprocessing transformation
            using the ``states``.
            If ``False``, assume the transformer is already calibrated.
        fit_basis : bool
            If ``True``, calibrate the high-to-low dimensional mapping
            using the ``states``.
            If ``False``, assume the basis is already calibrated.

        Returns
        -------
        states : list of s (r, k_i) ndarrays
            State snapshots in the reduced state space.
        lhs : list of s (r, k_i) ndarrays or None
            Left-hand side regression data in the reduced state space.
        inputs : list of s (m, k_i) ndarrays or None
            Processed input training data.
        """
        # Lifting.
        if self.lifter is not None:
            if lhs is not None:
                if self._iscontinuous:
                    lhs = [
                        self.lifter.lift_ddts(Q, Z)
                        for Q, Z in zip(states, lhs)
                    ]
                else:
                    lhs = [self.lifter.lift(Z) for Z in lhs]
            states = [self.lifter.lift(Q) for Q in states]

        # Preprocessing.
        if self.transformer is not None:
            if fit_transformer:
                self.transformer.fit(np.hstack(states))
            states = [self.transformer.transform(Q) for Q in states]
            if lhs is not None:
                if self._iscontinuous:
                    lhs = [self.transformer.transform_ddts(Z) for Z in lhs]
                else:
                    lhs = [self.transformer.transform(Z) for Z in lhs]

        # Dimensionality reduction.
        if self.basis is not None:
            if fit_basis:
                self.basis.fit(np.hstack(states))
            states = [self.basis.compress(Q) for Q in states]
            if lhs is not None:
                lhs = [self.basis.compress(Z) for Z in lhs]

        # Time derivative estimation / discrete LHS
        if lhs is None:
            if self._iscontinuous:
                if inputs is None:
                    states, lhs = zip(
                        *[self.ddt_estimator.estimate(Q) for Q in states]
                    )
                else:
                    states, lhs, inputs = zip(
                        *[
                            self.ddt_estimator.estimate(Q, U)
                            for Q, U in zip(states, inputs)
                        ]
                    )
            else:
                lhs = [Q[:, 1:] for Q in states]
                states = [Q[:, :-1] for Q in states]
                if inputs is not None:
                    inputs = [
                        U[..., : Q.shape[1]] for Q, U in zip(states, inputs)
                    ]

        return states, lhs, inputs

    def _process_test_cases(self, test_cases, TestCaseClass):
        if test_cases is not None:
            if isinstance(test_cases, TestCaseClass):
                test_cases = [test_cases]
            processed_test_cases = []
            for tcase in test_cases:
                if not isinstance(tcase, TestCaseClass):
                    raise TypeError(
                        "test cases must be "
                        f"'utils.{TestCaseClass.__name__}' objects"
                    )
                processed_test_cases.append(
                    tcase.copy(self.encode(tcase.initial_conditions))
                )
            return processed_test_cases
        return []

    def _get_stability_limits(self, states, stability_margin):
        shifts = [np.mean(Q, axis=1).reshape((-1, 1)) for Q in states]
        limits = [
            stability_margin * np.abs(Q - qbar).max()
            for Q, qbar in zip(states, shifts)
        ]
        for ell, lim in enumerate(limits):
            if lim == 0:  # pragma: no cover
                warnings.warn(
                    "ignoring stability limit"
                    f" for constant training trajectory {ell}",
                    errors.OpInfWarning,
                )
                limits[ell] = np.inf
        return shifts, limits

    def _fit_and_return_training_data(
        self,
        parameters,
        states,
        lhs,
        inputs,
        fit_transformer,
        fit_basis,
    ):
        """Process the training data, fit the model, and return the processed
        training data.
        """
        self._check_fit_args(lhs=lhs, inputs=inputs)
        if parameters is None:
            states, lhs, inputs = self._fix_single_trajectory(
                states, lhs, inputs
            )

        states, lhs, inputs = self._process_training_data(
            states=states,
            lhs=lhs,
            inputs=inputs,
            fit_transformer=fit_transformer,
            fit_basis=fit_basis,
        )

        if parameters is None:
            inputdata = None if inputs is None else np.hstack(inputs)
            self.model.fit(np.hstack(states), np.hstack(lhs), inputdata)
        else:
            self.model.fit(parameters, states, lhs, inputs)

        return states

    def fit_regselect_continuous(
        self,
        candidates: list,
        train_time_domains: list,
        parameters: list,
        states: list,
        ddts: list = None,
        input_functions: list = None,
        fit_transformer: bool = True,
        fit_basis: bool = True,
        regularizer_factory=None,
        gridsearch_only: bool = False,
        test_time_length: float = 0,
        stability_margin: float = 5.0,
        test_cases: list = None,
        verbose: bool = False,
        **predict_options: dict,
    ):
        """Calibrate the time-continuous model to training data, selecting the
        regularization hyperparameter(s) that minimize the training error while
        maintaining stability over the testing regime.

        This method requires the :attr:`model` to be time-continuous (use
        :meth:`fit_regselect_discrete` for fully discrete models) and to have a
        ``solver`` of one of the following types:

        * :class:`opinf.lstsq.L2Solver`
        * :class:`opinf.lstsq.L2DecoupledSolver`
        * :class:`opinf.lstsq.TikhonovSolver`
        * :class:`opinf.lstsq.TikhonovDecoupledSolver`

        The ``model.solver.regularizer`` is repeatedly adjusted, and the model
        is recalibrated, until a best regularization is selected.
        See :cite:`mcquarrie2021combustion,mcquarrie2023parametric`.

        Parameters
        ----------
        candidates : list of regularization hyperparameters
            Regularization hyperparameters to check before carrying out a
            derivative-free optimization.
        train_time_domains : list of s (k_i,) ndarrays
            Time domain corresponding to the training states.
        parameters : list of s (floats or (p,) ndarrays)
            Parameter values for which training data are available.
        states : list of s (n, k_i) ndarrays
            State snapshots in the original state space. Each array
            ``states[i]`` is the data corresponding to parameter value
            ``parameters[i]``; each column ``states[i][:, j]`` is one snapshot.
        ddts : list of s (n, k_i) ndarrays or None
            Snapshot time derivative data. Each array ``ddts[i]`` are the time
            derivatives of ``states[i]``; each column ``ddts[i][:, j]``
            corresponds to the snapshot ``states[i][:, j]``. If ``None``
            (default), these are estimated using :attr:`ddt_estimator`.
        input_functions : list of s callables or None
            Input functions mapping time to input vectors. Only required if the
            :attr:`model` takes external inputs. Each ``input_functions[i]``
            is the function corresponding to ``states[i]``, and
            ``input_functions[i](train_time_domains[i][j])`` is the input
            vector corresponding to the snapshot ``states[i][:, j]``.
        fit_transformer : bool
            If ``True`` (default), calibrate the preprocessing transformation
            using the ``states``.
            If ``False``, assume the transformer is already calibrated.
        fit_basis : bool
            If ``True`` (default), calibrate the high-to-low dimensional
            mapping using the ``states``.
            If ``False``, assume the basis is already calibrated.
        regularizer_factory : callable or None
            Function mapping regularization hyperparameters to the full
            regularizer. Specifically, ``regularizer_factory(candidates[i])``
            will be assigned to ``model.solver.regularizer`` for each ``i``.
            If ``None`` (default), set ``regularizer_factory()`` to the
            identity function.
        gridsearch_only : bool
            If ``True``, stop after checking all regularization ``candidates``
            and do not follow up with optimization.
        test_time_length : float or None
            Amount of time after the training regime in which to require model
            stability.
        stability_margin : float
            Factor by which the predicted reduced states may deviate from the
            range of the training reduced states without the trajectory being
            classified as unstable.
        test_cases : list of ContinuousRegTest objects
            Additional test cases for which the model is required to be stable.
            See :class:`opinf.utils.ContinuousRegTest`.
        verbose : bool
            If ``True``, print information during the regularization selection.
        predict_options : dict or None
            Extra arguments for :meth:`opinf.models.ContinuousModel.predict`,
            for example, ``method="BDF"``.

        Returns
        -------
        self

        Notes
        -----
        The ``train_time_domains`` may be a single one-dimensional array, in
        which case it is assumed that each trajectory ``states[i]`` corresponds
        to the same time domain. Similarly, if ``input_functions`` is a single
        callable, it is assumed to be the input function for each trajectory.
        """
        if not self._iscontinuous:
            raise AttributeError(
                "this method is for time-continuous models only, "
                "use fit_regselect_discrete()"
            )
        if not hasattr(self.model, "solver") or not hasattr(
            self.model.solver, "regularizer"
        ):
            raise AttributeError(
                "this method requires a model with a 'solver' attribute "
                "which has a 'regularizer' attribute"
            )

        if parameters is None:
            states, ddts, input_functions = self._fix_single_trajectory(
                states, ddts, input_functions
            )

        # Validate arguments.
        if np.isscalar(train_time_domains[0]):
            train_time_domains = [train_time_domains] * len(states)
        for t, Q in zip(train_time_domains, states):
            if t.shape != (Q.shape[1],):
                raise errors.DimensionalityError(
                    "train_time_domains and states not aligned"
                )
        if input_functions is not None:
            if callable(input_functions):  # one global input function.
                input_functions = [input_functions] * len(states)
            if not callable(input_functions[0]):
                raise TypeError(
                    "argument 'input_functions' must be sequence of callables"
                )
            inputs = [  # evaluate the inputs over the time domain.
                np.column_stack([u(tt) for tt in t])
                for u, t in zip(input_functions, train_time_domains)
            ]
        else:
            inputs = None
        if test_time_length < 0:
            raise ValueError("argument 'test_time_length' must be nonnegative")
        if regularizer_factory is None:
            regularizer_factory = _identity
        processed_test_cases = self._process_test_cases(
            test_cases, utils.ContinuousRegTest
        )

        # Fit the model for the first time.
        states = self._fit_and_return_training_data(
            parameters=parameters,
            states=states,
            lhs=ddts,
            inputs=inputs,
            fit_transformer=fit_transformer,
            fit_basis=fit_basis,
        )

        # Set up the regularization selection.
        shifts, limits = self._get_stability_limits(states, stability_margin)

        def unstable(_Q, ell, size):
            """Return ``True`` if the solution is unstable."""
            if _Q.shape[-1] != size:
                return True
            return np.abs(_Q - shifts[ell]).max() > limits[ell]

        # Extend the training time domains by the testing time length.
        if test_time_length > 0:
            time_domains = []
            for t_train in train_time_domains:
                dt = np.mean(np.diff(t_train))
                t_test = t_train[-1] + np.linspace(
                    dt,
                    dt + test_time_length,
                    int(test_time_length / dt),
                )
                time_domains.append(np.concatenate(((t_train, t_test))))
        else:
            time_domains = train_time_domains

        if input_functions is None:
            input_functions = [None] * len(states)
        loop_collections = [states, input_functions, time_domains]
        if is_parametric := parameters is not None:
            loop_collections.insert(0, parameters)

        def update_model(reg_params):
            """Reset the regularizer and refit the model operators."""
            self.model.solver.regularizer = regularizer_factory(reg_params)
            self.model.refit()

        def training_error(reg_params):
            """Compute the training error for a single regularization
            candidate by solving the model, checking for stability, and
            comparing to available training data.
            """
            update_model(reg_params)

            # Pass stability checks.
            for tcase in processed_test_cases:
                if not tcase.evaluate(self.model, **predict_options):
                    return np.inf

            # Compute training error.
            error = 0
            for ell, entries in enumerate(zip(*loop_collections)):
                if is_parametric:
                    params, Q, input_func, t = entries
                    predict_args = (params, Q[:, 0], t, input_func)
                else:
                    Q, input_func, t = entries
                    predict_args = (Q[:, 0], t, input_func)
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    solution = self.model.predict(
                        *predict_args, **predict_options
                    )
                if unstable(solution, ell, t.size):
                    return np.inf
                trainsize = Q.shape[-1]
                solution_train = solution[:, :trainsize]
                error += post.Lp_error(Q, solution_train, t[:trainsize])[1]
            return error / len(states)

        # BUG: training states may not align with the first `trainsize`
        # entries of `solution` because `Q` (from `states`) may have been
        # cropped by the time derivative estimator.
        # This is not an issue for fully discrete models.
        # Potential fix: implement some kind of mask in ddt_estimator and
        # replace `solution[:, :trainsize]` -> `solution[:, the_ddt_mask]`.

        best_regularization = utils.gridsearch(
            training_error,
            candidates,
            gridsearch_only=gridsearch_only,
            label="regularization",
            verbose=verbose,
        )

        update_model(best_regularization)
        return self

    def fit_regselect_discrete(
        self,
        candidates: list,
        parameters: list,
        states: list,
        inputs: list = None,
        fit_transformer: bool = True,
        fit_basis: bool = True,
        regularizer_factory=None,
        gridsearch_only: bool = False,
        num_test_iters: int = 0,
        stability_margin: float = 5.0,
        test_cases: list = None,
        verbose: bool = False,
    ):
        """Calibrate the fully discrete model to training data, selecting the
        regularization hyperparameter(s) that minimize the training error
        while maintaining stability over the testing regime.

        This method requires the :attr:`model` to be fully discrete (use
        :meth:`fit_regselect_continuous` for time-continuous models) and to
        have a ``solver`` of one of the following types:

        * :class:`opinf.lstsq.L2Solver`
        * :class:`opinf.lstsq.L2DecoupledSolver`
        * :class:`opinf.lstsq.TikhonovSolver`
        * :class:`opinf.lstsq.TikhonovDecoupledSolver`

        The ``model.solver.regularizer`` is repeatedly adjusted, and the model
        is recalibrated, until a best regularization is selected.
        See :cite:`mcquarrie2021combustion,mcquarrie2023parametric`.

        Parameters
        ----------
        candidates : list of regularization hyperparameters
            Regularization hyperparameters to check. If a single hyperparameter
            is given, use it as the start of an optimization-based search.
        parameters : list of s (floats or (p,) ndarrays)
            Parameter values for which training data are available.
        states : list of s (n, k_i) ndarrays
            State snapshots in the original state space. Each array
            ``states[i]`` is the data corresponding to parameter value
            ``parameters[i]``; each column ``states[i][:, j]`` is one snapshot.
            This method assumes the snapshots are sequential, i.e., the model
            maps ``states[i][:, j]`` to ``states[i][:, j+1]``.
        inputs : list of s (m, k_i + num_test_iters) ndarrays
            Inputs corresponding to the training data, together with inputs
            for the testing regime. Only required if the :attr:`model` takes
            external inputs.
        fit_transformer : bool
            If ``True`` (default), calibrate the preprocessing transformation
            using the ``states``.
            If ``False``, assume the transformer is already calibrated.
        fit_basis : bool
            If ``True`` (default), calibrate the high-to-low dimensional
            mapping using the ``states``.
            If ``False``, assume the basis is already calibrated.
        regularizer_factory : callable or None
            Function mapping regularization hyperparameters to the full
            regularizer. Specifically, ``regularizer_factory(candidates[i])``
            will be assigned to ``model.solver.regularizer`` for each ``i``.
        gridsearch_only : bool
            If ``True``, stop after checking all regularization ``candidates``
            and do not follow up with optimization.
        num_test_iters : int
            Number of iterations after the training data in which to require
            model stability.
        stability_margin : float,
            Factor by which the reduced states may deviate from the range of
            the training data without being flagged as unstable.
        test_cases : list of DiscreteRegTest objects
            Additional test cases for which the model is required to be stable.
            See :class:`opinf.utils.DiscreteRegTest`.
        verbose : bool
            If ``True``, print information during the regularization selection.
        """
        if self._iscontinuous:
            raise AttributeError(
                "this method is for fully discrete models only, "
                "use fit_regselect_continuous()"
            )
        if not hasattr(self.model, "solver") or not hasattr(
            self.model.solver, "regularizer"
        ):
            raise AttributeError(
                "this method requires a model with a 'solver' attribute "
                "which has a 'regularizer' attribute"
            )

        if parameters is None:
            states, _, inputs = self._fix_single_trajectory(
                states, None, inputs
            )

        # Validate arguments.
        if num_test_iters < 0:
            raise ValueError(
                "argument 'num_test_iters' must be a nonnegative integer"
            )
        if inputs is not None:
            if len(inputs) != len(states):
                raise errors.DimensionalityError(
                    f"{len(states)} state trajectories but "
                    f"{len(inputs)} input trajectories detected"
                )
            for Q, U in zip(states, inputs):
                if U.shape[-1] < Q.shape[1] + num_test_iters:
                    raise ValueError(
                        "argument 'inputs' must contain enough data for "
                        f"{num_test_iters} iterations after the training data"
                    )
        if regularizer_factory is None:
            regularizer_factory = _identity
        processed_test_cases = self._process_test_cases(
            test_cases, utils.DiscreteRegTest
        )

        # Fit the model for the first time.
        states = self._fit_and_return_training_data(
            parameters=parameters,
            states=states,
            lhs=None,
            inputs=inputs,
            fit_transformer=fit_transformer,
            fit_basis=fit_basis,
        )

        # Set up the regularization selection.
        shifts, limits = self._get_stability_limits(states, stability_margin)

        def unstable(_Q, ell):
            """Return ``True`` if the solution is unstable."""
            if np.isnan(_Q).any() or np.isinf(_Q).any():
                return True
            return np.any(np.abs(_Q - shifts[ell]).max() > limits[ell])

        # Extend the iteration counts by the number of testing iterations.
        num_iters = [Q.shape[-1] for Q in states]
        if num_test_iters > 0:
            num_iters = [n + num_test_iters for n in num_iters]

        if inputs is None:
            inputs = [None] * len(states)
        loop_collections = [states, inputs, num_iters]
        if is_parametric := parameters is not None:
            loop_collections.insert(0, parameters)

        def update_model(reg_params):
            """Reset the regularizer and refit the model operators."""
            self.model.solver.regularizer = regularizer_factory(reg_params)
            self.model.refit()

        def training_error(reg_params):
            """Compute the mean training error for a single regularization
            candidate by solving the model, checking for stability, and
            comparing to available training data.
            """
            update_model(reg_params)

            # Pass stability checks.
            for tcase in processed_test_cases:
                if not tcase.evaluate(self.model):
                    return np.inf

            # Solve the model, check for stability, and compute training error.
            error = 0
            for ell, entries in enumerate(zip(*loop_collections)):
                if is_parametric:
                    params, Q, U, niter = entries
                    predict_args = (params, Q[:, 0], niter, U)
                else:
                    Q, U, niter = entries
                    predict_args = (Q[:, 0], niter, U)
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    solution = self.model.predict(*predict_args)
                if unstable(solution, ell):
                    return np.inf
                error += post.frobenius_error(Q, solution[:, : Q.shape[-1]])[1]
            return error / len(states)

        best_regularization = utils.gridsearch(
            training_error,
            candidates,
            gridsearch_only=gridsearch_only,
            label="regularization",
            verbose=verbose,
        )

        update_model(best_regularization)
        return self

    # Abstracts ---------------------------------------------------------------
    @abc.abstractmethod
    def fit(self, *args, **kwargs):
        """Calibrate the model to training data."""
        pass  # pragma: no cover

    @abc.abstractmethod
    def predict(self, *args, **kwargs):
        """Evaluate the model."""
        raise NotImplementedError  # pragma: no cover

    # Verification ------------------------------------------------------------
    # TODO



# utils/_hdf5.py
"""Utilities for HDF5 file interaction."""

__all__ = [
    "hdf5_savehandle",
    "hdf5_loadhandle",
    "save_sparray",
    "load_sparray",
]

import os
import h5py
import warnings
import scipy.sparse as sparse

from .. import errors


# File handle classes =========================================================
class _hdf5_filehandle:
    """Get a handle to an open HDF5 file to read or write to.

    Parameters
    ----------
    filename : str or h5py File/Group handle
        * str : Name of the file to interact with.
        * h5py File/Group handle : handle to part of an already open HDF5 file.
    mode : str
        Type of interaction for the HDF5 file.
        * "save" : Open the file for writing only.
        * "load" : Open the file for reading only.
    overwrite : bool
        If True, overwrite the file if it already exists. If False,
        raise a FileExistsError if the file already exists.
        Only applies when ``mode = "save"``.
    """

    def __init__(self, filename, mode, overwrite=False):
        """Open the file handle."""
        if isinstance(filename, h5py.HLObject):
            # `filename` is already an open HDF5 file.
            self.file_handle = filename
            self.close_when_done = False

        elif mode == "save":
            # `filename` is the name of a file to create for writing.
            if not filename.endswith(".h5"):
                warnings.warn(
                    "expected file with extension '.h5'",
                    errors.OpInfWarning,
                )
            if os.path.isfile(filename) and not overwrite:
                raise FileExistsError(f"{filename} (overwrite=True to ignore)")
            self.file_handle = h5py.File(filename, "w")
            self.close_when_done = True

        elif mode == "load":
            # `filename` is the name of an existing file to read from.
            if not os.path.isfile(filename):
                raise FileNotFoundError(filename)
            self.file_handle = h5py.File(filename, "r")
            self.close_when_done = True

        else:
            raise ValueError(f"invalid mode '{mode}'")

    def __enter__(self):
        """Return the handle to the open HDF5 file."""
        return self.file_handle

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Close the file if needed."""
        if self.close_when_done:
            self.file_handle.close()
        if exc_type:
            raise


class hdf5_savehandle(_hdf5_filehandle):
    """Get a handle to an open HDF5 file to write to.

    Parameters
    ----------
    savefile : str or h5py File/Group handle
        * str : Name of the file to save to.
        * h5py File/Group handle : handle to part of an already open HDF5 file
          to save data to.
    overwrite : bool
        If ``True``, overwrite the file if it already exists.
        If ``False``, raise a ``FileExistsError`` if the file already exists.

    Examples
    --------
    >>> with hdf5_savehandle("file_to_save_to.h5", False) as hf:
    ...     hf.create_dataset("dataset_label", data=dataset_to_save)
    """

    def __init__(self, savefile, overwrite):
        return _hdf5_filehandle.__init__(self, savefile, "save", overwrite)


class hdf5_loadhandle(_hdf5_filehandle):
    """Get a handle to an open HDF5 file to read from.

    Parameters
    ----------
    loadfile : str or h5py File/Group handle
        * str : Name of the file to read from.
        * h5py File/Group handle : handle to part of an already open HDF5 file
          to read data from.

    Examples
    --------
    >>> with hdf5_loadhandle("file_to_read_from.h5") as hf:
    ...    dataset = hf["dataset_label"][:]
    """

    def __init__(self, loadfile):
        return _hdf5_filehandle.__init__(self, loadfile, "load")

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Close the file if needed. Raise a LoadfileFormatError if needed."""
        try:
            _hdf5_filehandle.__exit__(self, exc_type, exc_value, exc_traceback)
        except errors.LoadfileFormatError:
            raise
        except Exception as ex:
            raise errors.LoadfileFormatError(ex.args[0]) from ex


# Other tools =================================================================
def save_sparray(group: h5py.Group, arr: sparse.sparray) -> None:
    """Save a :mod:`scipy.sparse` matrix efficiently in an HDF5 group.

    This method mimics the behavior of :meth:`scipy.sparse.save_npz()` but
    for an open HDF5 file. See :func:`load_sparray()`.

    Parameters
    ----------
    arr : scipy.sparse.sparray
        Sparse SciPy array, in any sparse format.
    group : h5py.Group
        HDF5 group to save the sparse array to.

    Examples
    --------
    >>> import h5py
    >>> import scipy.sparse as sparse
    >>> from opinf.utils import save_sparray, load_sparray

    # Create a sparse array.
    >>> A = sparse.dok_array((100, 100), dtype=float)
    >>> A[0, 5] = 12
    >>> A[4, 1] = 123.456
    >>> A
    <100x100 sparse array of type '<class 'numpy.float64'>'
        with 2 stored elements in Dictionary Of Keys format>
    >>> print(A)
      (np.int32(0), np.int32(5))    12.0
      (np.int32(4), np.int32(1))    123.456

    # Save the sparse array to an HDF5 file.
    >>> with h5py.File("myfile.h5", "w") as hf:
    ...     save_sparray(hf.create_group("sparsearray"), A)

    # Load the sparse array from the file.
    >>> with h5py.File("myfile.h5", "r") as hf:
    ...     B = load_sparray(hf["sparsearray"])
    >>> B
    <100x100 sparse array of type '<class 'numpy.float64'>'
        with 2 stored elements in Dictionary Of Keys format>
    >>> print(B)
      (np.int32(0), np.int32(5))    12.0
      (np.int32(4), np.int32(1))    123.456
    """
    if not sparse.issparse(arr):
        raise TypeError("second arg must be a scipy.sparse array")

    # Convert to COO format and save data attributes.
    A = arr.tocoo()
    group.create_dataset("data", data=A.data)
    group.create_dataset("row", data=A.row)
    group.create_dataset("col", data=A.col)
    group.attrs["shape"] = A.shape
    group.attrs["arrtype"] = type(arr).__name__[:3]


def load_sparray(group: h5py.Group) -> sparse.sparray:
    """Save a :mod:`scipy.sparse` matrix efficiently in an HDF5 group.

    This method mimics the behavior of :meth:`scipy.sparse.load_npz()` but
    for an open HDF5 file. See :func:`save_sparray()`.

    Parameters
    ----------
    group : h5py.Group
        HDF5 group create and save the sparse array to.

    Returns
    -------
    arr : scipy.sparse.sparray
        Sparse SciPy array, in the sparse format it was in before saving.

    Examples
    --------
    >>> import h5py
    >>> import scipy.sparse as sparse
    >>> from opinf.utils import save_sparray, load_sparray

    # Create a sparse array.
    >>> A = sparse.dok_array((100, 100), dtype=float)
    >>> A[0, 5] = 12
    >>> A[4, 1] = 123.456
    >>> A
    <100x100 sparse array of type '<class 'numpy.float64'>'
        with 2 stored elements in Dictionary Of Keys format>
    >>> print(A)
      (np.int32(0), np.int32(5))    12.0
      (np.int32(4), np.int32(1))    123.456

    # Save the sparse array to an HDF5 file.
    >>> with h5py.File("myfile.h5", "w") as hf:
    ...     save_sparray(hf.create_group("sparsearray"), A)

    # Load the sparse array from the file.
    >>> with h5py.File("myfile.h5", "r") as hf:
    ...     B = load_sparray(hf["sparsearray"])
    >>> B
    <100x100 sparse array of type '<class 'numpy.float64'>'
        with 2 stored elements in Dictionary Of Keys format>
    >>> print(B)
      (np.int32(0), np.int32(5))    12.0
      (np.int32(4), np.int32(1))    123.456
    """
    A = sparse.coo_matrix(
        (group["data"], (group["row"], group["col"])),
        shape=group.attrs["shape"],
    )
    arrtype = str(group.attrs["arrtype"])
    return getattr(A, f"to{arrtype}")()



# models/mono/_parametric.py
"""Parametric monolithic dynamical systems models."""

__all__ = [
    "ParametricDiscreteModel",
    "ParametricContinuousModel",
    "InterpDiscreteModel",
    "InterpContinuousModel",
    # Deprecations:
    "InterpolatedDiscreteModel",
    "InterpolatedContinuousModel",
]

import warnings
import numpy as np
import scipy.interpolate as spinterpolate

from ._base import _Model
from ._nonparametric import (
    _FrozenDiscreteModel,
    _FrozenContinuousModel,
)
from ... import errors, utils, operators as _operators
from ...operators import _utils as oputils


# Base classes ================================================================
class _ParametricModel(_Model):
    """Base class for parametric monolithic models."""

    _ModelClass = NotImplemented  # Must be specified by child classes.

    # ModelClass properties ---------------------------------------------------
    @property
    def _LHS_ARGNAME(self):  # pragma: no cover
        """Name of LHS argument in fit(), e.g., "ddts"."""
        return self._ModelClass._LHS_ARGNAME

    @property
    def _LHS_LABEL(self):  # pragma: no cover
        """String representation of LHS, e.g., "dq / dt"."""
        return self._ModelClass._LHS_LABEL

    @property
    def _STATE_LABEL(self):  # pragma: no cover
        """String representation of state, e.g., "q(t)"."""
        return self._ModelClass._STATE_LABEL

    @property
    def _INPUT_LABEL(self):  # pragma: no cover
        """String representation of input, e.g., "u(t)"."""
        return self._ModelClass._INPUT_LABEL

    # Properties: operators ---------------------------------------------------
    _operator_abbreviations = dict()

    def _isvalidoperator(self, op):
        """All monolithic operators are allowed."""
        return isinstance(
            op,
            (
                _operators.OperatorTemplate,
                _operators.ParametricOperatorTemplate,
            ),
        )

    @staticmethod
    def _check_operator_types_unique(ops):
        """Raise a ValueError if any two operators represent the same kind
        of operation (e.g., two constant operators).
        """
        OpClasses = {
            (op._OperatorClass if oputils.is_parametric(op) else type(op))
            for op in ops
        }
        if len(OpClasses) != len(ops):
            raise ValueError("duplicate type in list of operators to infer")

    def _get_operator_of_type(self, OpClass):
        """Return the first operator in the model corresponding to the
        operator class ``OpClass``.
        """
        for op in self.operators:
            if oputils.is_parametric(op) and op._OperatorClass is OpClass:
                return op
            if oputils.is_nonparametric(op) and isinstance(op, OpClass):
                return op

    @property
    def operators(self):
        """Operators comprising the terms of the model."""
        return _Model.operators.fget(self)

    @operators.setter
    def operators(self, ops):
        """Set the operators."""
        _Model.operators.fset(self, ops)

        # Check at least one operator is parametric.
        parametric_operators = [
            op for op in self.operators if oputils.is_parametric(op)
        ]
        if len(parametric_operators) == 0:
            warnings.warn(
                "no parametric operators detected, "
                "consider using a nonparametric model class",
                errors.OpInfWarning,
            )

        # Check that not every operator is interpolated.
        if not isinstance(self, _InterpModel):
            interpolated_operators = [
                op for op in self.operators if oputils.is_interpolated(op)
            ]
            if len(interpolated_operators) == len(self.operators):
                warnings.warn(
                    "all operators interpolatory, "
                    "consider using an InterpModel class",
                    errors.OpInfWarning,
                )
        self._synchronize_parameter_dimensions()

    # Properties: dimensions --------------------------------------------------
    @property
    def parameter_dimension(self):
        r"""Dimension :math:`p` of a parameter vector :math:`\bfmu`."""
        return self.__p

    def _synchronize_parameter_dimensions(self, newdim=None):
        """Synchronize the parameter_dimension attribute for each operator."""
        # Get any non-None parameter dimensions and check for uniqueness.
        ps = {
            op.parameter_dimension
            for op in self.operators
            if oputils.is_parametric(op) and op.parameter_dimension is not None
        }
        if len(ps) > 1:
            raise errors.DimensionalityError(
                "operators not aligned "
                "(parameter_dimension must be the same for all operators)"
            )
        p = ps.pop() if len(ps) == 1 else None

        # Check operator parameter_dimension matches new parameter_dimension.
        if newdim is not None:
            if p is None:
                p = newdim
            if p != newdim:
                raise errors.DimensionalityError(
                    f"{p} = each operator.parameter_dimension != "
                    f"parameter dimension = {newdim}"
                )

        # Ensure all parametric operators have the same parameter_dimension.
        if p is not None:
            for op in self.operators:
                if (
                    oputils.is_parametric(op)
                    and op.parameter_dimension is None
                ):
                    op.parameter_dimension = p

        # Set the model's parameter_dimension to the same as the operators.
        self.__p = p

    # Fitting -----------------------------------------------------------------
    def _process_fit_arguments(self, parameters, states, lhs, inputs):
        """Prepare training data for Operator Inference by extracting
        dimensions, validating data sizes, and modifying the left-hand side
        data if there are any known operators.
        """
        # Clear non-intrusive operator data.
        self._clear()

        # Process parameters.
        if (dim := len(shape := np.shape(parameters))) == 1:
            p = 1
        elif dim == 2:
            p = shape[1]
        else:
            raise errors.DimensionalityError(
                "'parameters' must be a sequence of scalars or 1D arrays"
            )
        self._synchronize_parameter_dimensions(p)
        n_datasets = len(parameters)

        def _check_valid_dimension0(dataset, label):
            """Dimension 0 must be s (number of training parameters)."""
            if (datalen := len(dataset)) != n_datasets:
                raise errors.DimensionalityError(
                    f"len({label}) = {datalen} "
                    f"!= {n_datasets} = len(parameters)"
                )

        def _check_valid_dimension1(dataset, label):
            """Dimension 1 must be r (state dimensions)."""
            for i, subset in enumerate(dataset):
                if (dim := len(subset)) != (r := self.state_dimension):
                    raise errors.DimensionalityError(
                        f"len({label}[{i}]) = {dim} != {r} = r"
                    )

        # Process states, extract model dimension if needed.
        states = [np.atleast_2d(Q) for Q in states]
        if self.state_dimension is None:
            self.state_dimension = states[0].shape[0]
        _check_valid_dimension0(states, "states")
        _check_valid_dimension1(states, "states")

        def _check_valid_dimension2(dataset, label):
            """Dimension 2 must match across datasets (number of snapshots)."""
            for i, subset in enumerate(dataset):
                if (dim := subset.shape[-1]) != (k := states[i].shape[-1]):
                    raise errors.DimensionalityError(
                        f"{label}[{i}].shape[-1] = {dim} "
                        f"!= {k} = states[{i}].shape[-1]"
                    )

        # Process LHS.
        lhs = [np.atleast_2d(L) for L in lhs]
        _check_valid_dimension0(lhs, self._LHS_ARGNAME)
        _check_valid_dimension1(lhs, self._LHS_ARGNAME)
        _check_valid_dimension2(lhs, self._LHS_ARGNAME)

        # Process inputs, extract input dimension if needed.
        self._check_inputargs(inputs, "inputs")
        if self._has_inputs:
            inputs = [np.atleast_2d(U) for U in inputs]
            if not self.input_dimension:
                self.input_dimension = inputs[0].shape[0]
            _check_valid_dimension0(inputs, "inputs")
            for i, subset in enumerate(inputs):
                if (dim := subset.shape[0]) != (m := self.input_dimension):
                    raise errors.DimensionalityError(
                        f"inputs[{i}].shape[0] = {dim} != {m} = m"
                    )
            _check_valid_dimension2(inputs, "inputs")
        elif inputs is None:
            inputs = [None] * n_datasets

        # Subtract known operator evaluations from the LHS.
        for ell in self._indices_of_known_operators:
            op = self.operators[ell]
            _isparametric = oputils.is_parametric(op)
            for i, lhsi in enumerate(lhs):
                _args = [states[i], inputs[i]]
                if _isparametric:
                    _args.insert(0, parameters[i])
                lhs[i] = lhsi - op.apply(*_args)

        return parameters, states, lhs, inputs

    def _assemble_data_matrix(self, parameters, states, inputs):
        """Assemble the data matrix for operator inference."""
        blocks = []
        for i in self._indices_of_operators_to_infer:
            op = self.operators[i]
            if not oputils.is_parametric(op):
                block = np.hstack(
                    [op.datablock(Q, U) for Q, U in zip(states, inputs)]
                )
            else:
                block = op.datablock(parameters, states, inputs)
            blocks.append(block.T)
        return np.hstack(blocks)

    def _fit_solver(self, parameters, states, lhs, inputs=None):
        """Construct a solver for the operator inference least-squares
        regression."""
        (
            parameters_,
            states_,
            lhs_,
            inputs_,
        ) = self._process_fit_arguments(parameters, states, lhs, inputs)

        # Set training_parameters for interpolatory operators.
        for op in self.operators:
            if oputils.is_interpolated(op):
                op.set_training_parameters(parameters_)

        # Set up non-intrusive learning.
        D = self._assemble_data_matrix(parameters_, states_, inputs_)
        self.solver.fit(D, np.hstack(lhs_))
        self.__s = len(parameters_)

    def _extract_operators(self, Ohat):
        """Unpack the operator matrix and populate operator entries."""
        index = 0
        for i in self._indices_of_operators_to_infer:
            op = self.operators[i]
            if oputils.is_parametric(op):
                endex = index + op.operator_dimension(
                    self.__s, self.state_dimension, self.input_dimension
                )
                op.set_entries(Ohat[:, index:endex], fromblock=True)
            else:
                endex = index + op.operator_dimension(
                    self.state_dimension, self.input_dimension
                )
                op.set_entries(Ohat[:, index:endex])
            index = endex

    def refit(self):
        """Solve the Operator Inference regression using the data from the
        last :meth:`fit()` call, then extract the inferred operators.

        This method is useful for calibrating the model operators with
        different ``solver`` hyperparameters without reprocessing any training
        data. For example, if ``solver`` is an :class:`opinf.lstsq.L2Solver`,
        changing its ``regularizer`` attribute and calling this method solves
        the regression with the new regression value without re-factorizing the
        data matrix.
        """
        if self._fully_intrusive:
            warnings.warn(
                "all operators initialized explicitly, nothing to learn",
                errors.OpInfWarning,
            )
            return self

        # Execute non-intrusive learning.
        self._extract_operators(self.solver.solve())
        return self

    def fit(self, parameters, states, lhs, inputs=None):
        r"""Learn the model operators from data.

        The operators are inferred by solving the regression problem

        .. math::
           \min_{\Ophat}
           \sum_{i=1}^{s}\sum_{j=0}^{k_{i}-1}\left\|
           \Ophat(\qhat_{i,j}, \u_{i,j}; \bfmu_i) - \dot{\qhat}_{i,j}
           \right\|_2^2

        where
        :math:`\zhat = \Ophat(\qhat, \u)` is the model and

        * :math:`\qhat_j\in\RR^r` is a measurement of the state,
        * :math:`\u_j\in\RR^m` is a measurement of the input, and
        * :math:`\zhat_j\in\RR^r` is a measurement of the left-hand side
          of the model.

        The *operator matrix* :math:`\Ohat\in\RR^{r\times d(r,m)}` is such that
        :math:`\Ophat(\q,\u) = \Ohat\d(\qhat,\u)` for some data vector
        :math:`\d(\qhat,\u)\in\RR^{d(r,m)}`; the *data matrix*
        :math:`\D\in\RR^{k\times d(r,m)}` is given by
        :math:`[~\d(\qhat_0,\u_0)~~\cdots~~\d(\qhat_{k-1},\u_{k-1})~]\trp`.
        Finally,
        :math:`\Zhat = [~\zhat_0~~\cdots~~\zhat_{k-1}~]\in\RR^{r\times k}`.
        See the :mod:`opinf.operators` module for more explanation.

        The strategy for solving the regression, as well as any additional
        regularization or constraints, are specified by the ``solver`` in
        the constructor.

        Parameters
        ----------
        parameters : list of s (floats or (p,) ndarrays)
            Parameter values for which training data are available.
        states : list of s (r, k) ndarrays
            Snapshot training data. Each array ``states[i]`` is the data
            corresponding to parameter value ``parameters[i]``; each column
            ``states[i][:, j]`` is a single snapshot.
        lhs : list of s (r, k) ndarrays
            Left-hand side training data. Each array ``lhs[i]`` is the data
            corresponding to parameter value ``parameters[i]``; each column
            ``lhs[i][:, j]`` corresponds to the snapshot ``states[i][:, j]``.
            The interpretation of this argument depends on the setting:
            forcing data for steady-state problems, next iteration for
            discrete-time problems, and time derivatives of the state for
            continuous-time problems.
        inputs : list of s (m, k) or (k,) ndarrays, or None
            Input training data. Each array ``inputs[i]`` is the data
            corresponding to parameter value ``parameters[i]``; each column
            ``inputs[i][:, j]`` corresponds to the snapshot ``states[:, j]``.
            May be a two-dimensional array if :math:`m=1` (scalar input).

        Returns
        -------
        self
        """
        if self._fully_intrusive:
            warnings.warn(
                "all operators initialized explicitly, nothing to learn",
                errors.OpInfWarning,
            )
            return self

        self._fit_solver(parameters, states, lhs, inputs)
        return self.refit()

    # Parametric evaluation ---------------------------------------------------
    def evaluate(self, parameter):
        """Construct a nonparametric model by fixing the parameter value.

        Parameters
        ----------
        parameter : (p,) ndarray or float
            Parameter value at which to evaluate the model.

        Returns
        -------
        model : _NonparametricModel
            Nonparametric model of type ``ModelClass``.
        """
        return self._ModelClass(
            [
                op.evaluate(parameter) if oputils.is_parametric(op) else op
                for op in self.operators
            ]
        )

    def rhs(self, parameter, *args, **kwargs):
        r"""Evaluate the right-hand side of the model by applying each operator
        and summing the results.

        This is the function :math:`\Ophat(\qhat, \u; \bfmu)`
        where the model can be written as one of the following:

        * :math:`\ddt\qhat(t; \bfmu) = \Ophat(\qhat(t; \bfmu), \u(t); \bfmu)`
          (continuous time)
        * :math:`\qhat_{j+1}(\bfmu) = \Ophat(\qhat_j(\bfmu), \u_j; \bfmu)`
          (discrete time)
        * :math:`\hat{\mathbf{g}} = \Ophat(\qhat(\bmfu), \u; \bfmu)`
          (steady state)

        Parameters
        ----------
        parameter : (p,) ndarray
            Parameter value :math:`\bfmu`.
        args
            Positional arguments to ``ModelClass.rhs()``.
        kwargs
            Keyword arguments to ``ModelClass.rhs()``.

        Returns
        -------
        evaluation : (r,) ndarray
            Evaluation of the right-hand side of the model.

        Notes
        -----
        For repeated ``rhs()`` calls with the same parameter value, use
        :meth:`evaluate` to first get the nonparametric model corresponding
        to the parameter value.
        """
        return self.evaluate(parameter).rhs(*args, **kwargs)

    def jacobian(self, parameter, *args, **kwargs):
        r"""Construct and sum the state Jacobian of each model operator.

        This the derivative of the right-hand side of the model with respect
        to the state, i.e., the function
        :math:`\ddqhat\Ophat(\qhat, \u; \bmfu)`
        where the model can be written as one of the following:

        * :math:`\ddt\qhat(t; \bfmu) = \Ophat(\qhat(t; \bfmu), \u(t); \bfmu)`
          (continuous time)
        * :math:`\qhat(\bfmu)_{j+1} = \Ophat(\qhat(\bfmu)_{j}, \u_{j}; \bfmu)`
          (discrete time)
        * :math:`\hat{\mathbf{g}} = \Ophat(\qhat(\bfmu), \u; \bfmu)`
          (steady state)

        Parameters
        ----------
        parameter : (p,) ndarray
            Parameter value :math:`\bfmu`.
        args
            Positional arguments to ``ModelClass.jacobian()``.
        kwargs
            Keyword arguments to ``ModelClass.jacobian()``.

        Returns
        -------
        jac : (r, r) ndarray
            State Jacobian of the right-hand side of the model.

        Notes
        -----
        For repeated ``jacobian()`` calls with the same parameter value, use
        :meth:`evaluate` to first get the nonparametric model corresponding
        to the parameter value.
        """
        return self.evaluate(parameter).jacobian(*args, **kwargs)

    def predict(self, parameter, *args, **kwargs):
        r"""Solve the model at the given parameter value.

        Parameters
        ----------
        parameter : (p,) ndarray
            Parameter value :math:`\bfmu`.
        args
            Positional arguments to ``ModelClass.predict()``.
        kwargs
            Keyword arguments to ``ModelClass.predict()``.

        Notes
        -----
        For repeated ``predict()`` calls with the same parameter value, use
        :meth:`evaluate` to first get the nonparametric model corresponding
        to the parameter value.
        """
        return self.evaluate(parameter).predict(*args, **kwargs)


class _ParametricDiscreteMixin:
    """Mixin class for parametric models of discrete dynamical system."""

    _ModelClass = _FrozenDiscreteModel

    def fit(self, parameters, states, nextstates=None, inputs=None):
        r"""Learn the model operators from data.

        The operators are inferred by solving the regression problem

        .. math::
           \min_{\Ohat^{(1)},\ldots,\Ohat^{(s)}}
           \sum_{i=1}^{s}\sum_{j=0}^{k_{i}-1}\left\|
           \Ophat(\qhat_{i,j}, \u_{i,j}; \bfmu_i) - \zhat_{i,j}
           \right\|_2^2
           = \min_{\Ohat^{(1)},\ldots,\Ohat^{(s)}}
           \sum_{i=1}^{s}\left\|
           \D^{(i)}(\Ohat^{(i)})\trp
           - [~\zhat_{i,0}~~\cdots~~\zhat_{i,k_{i}-1}~]\trp
           \right\|_F^2

        where
        :math:`\zhat(\bfmu)_{j} = \Ophat(\qhat(\bfmu)_{j}, \u_{j}; \bfmu)
        = \Ohat(\bfmu)\d(\qhat(\bfmu)_{j}, \u_{j})` is the model and

        * :math:`\qhat_{i,j}\in\RR^r` is the :math:`j`-th measurement of the
          state corresponding to training parameter value :math:`\bfmu_i`,
        * :math:`\u_{i,j}\in\RR^m` is the :math:`j`-th measurement of the
          input corresponding to training parameter value :math:`\bfmu_i`,
        * :math:`\dot{qhat}_{i,j}\in\RR^r` is a measurement of the time
          derivative of the state corresponding to the state-input pair
          :math:`(\qhat_{i,j},\u_{i,j})`,
        * :math:`\Ohat^{(i)} = \Ohat(\bfmu)` is the operator matrix
          evaluated at training parameter value :math:`\bfmu_i`, and
        * :math:`\D^{(i)}` is the data matrix for data corresponding to
          training parameter value :math:`\bfmu_i`, given by
          :math:`[~\d(\qhat_{i,0},\u_{i,0})~~\cdots~~
          \d(\qhat_{i,k_i-1},\u_{i,k_i-1})~]\trp`.

        Because all operators in this model are interpolatory, the
        least-squares problem decouples into :math:`s` individual regressions.
        That is, for :math:`i = 1, \ldots, s`, we solve (independently) the
        regressions

        .. math::
           \min_{\Ohat^{(i)}}\left\|
           \D^{(i)}(\Ohat^{(i)})\trp
           - [~\zhat_{i,0}~~\cdots~~\zhat_{i,k_{i}-1}~]\trp
           \right\|_F^2

        and define the full operator matrix via elementwise interpolation,

        .. math::
           \Ohat(\bfmu) = \textrm{interpolate}(
           (\bfmu_1, \Ohat^{(i)}), \ldots, (\bfmu_s, \Ohat^{(s)}); \bfmu).

        The strategy for solving the regression, as well as any additional
        regularization or constraints, are specified by the ``solver`` in
        the constructor.

        Parameters
        ----------
        parameters : list of s scalars or (p,) 1D ndarrays
            Parameter values for which the operator entries are known
            or will be inferred from data.
        states : list of s (r, k_i) ndarrays
            Snapshot training data. Each array ``states[i]`` is the data
            corresponding to parameter value ``parameters[i]``; each column
            ``states[i][:, j]`` is a single snapshot.
        nextstates : list of s (r, k_i) ndarrays
            Next iteration training data. Each array ``nextstates[i]`` is the
            data corresponding to parameter value ``parameters[i]``; each
            column ``nextstates[i][:, j]`` is the iteration following
            ``states[i][:, j]``.
        inputs : list of s (m, k_i) or (k_i,) ndarrays, or None
            Input training data. Each array ``inputs[i]`` is the data
            corresponding to parameter value ``parameters[i]``; each column
            ``inputs[i][:, j]`` corresponds to the snapshot ``states[:, j]``.
            May be a two-dimensional array if :math:`m=1` (scalar input).

        Returns
        -------
        self
        """
        if nextstates is None:
            nextstates = [Q[:, 1:] for Q in states]
            states = [Q[:, :-1] for Q in states]
        if inputs is not None:
            inputs = [U[..., : Q.shape[1]] for U, Q in zip(inputs, states)]
        return super().fit(parameters, states, nextstates, inputs)

    def rhs(self, parameter, state, input_=None):
        r"""Evaluate the right-hand side of the model by applying each operator
        and summing the results.

        This is the function :math:`\fhat(\qhat, \u; \bfmu)`
        where the model is given by
        :math:`\qhat(\bfmu)_{j+1} = \fhat(\qhat(\bfmu)_{j}, \u_{j}; \bfmu)`.

        Parameters
        ----------
        parameter : (p,) ndarray
            Parameter value :math:`\bfmu`.
        state : (r,) ndarray
            State vector :math:`\qhat`.
        input_ : (m,) ndarray or None
            Input vector :math:`\u`.

        Returns
        -------
        nextstate : (r,) ndarray
            Evaluation of the right-hand side of the model.

        Notes
        -----
        For repeated ``rhs()`` calls with the same parameter value, use
        :meth:`evaluate` to first get the nonparametric model corresponding
        to the parameter value.

        .. code-block::

           # Instead of this...
           >>> values = [parametric_model.rhs(parameter, q, input_)
           ...           for q in list_of_states]
           # ...it is faster to do this.
           >>> model_at_parameter = parametric_model.evaluate(parameter)
           >>> values = [model_at_parameter.rhs(parameter, q, input_)
           ...           for q in list_of_states]
        """
        return super().rhs(parameter, state, input_)

    def jacobian(self, parameter, state, input_=None):
        r"""Sum the state Jacobian of each model operator.

        This the derivative of the right-hand side of the model with respect
        to the state, i.e., the function :math:`\ddqhat\fhat(\qhat, \u; \bfmu)`
        where the model is given by
        :math:`\qhat(\bfmu)_{j+1} = \fhat(\qhat(\bfmu)_{j}, \u_{j}; \bfmu)`.

        Parameters
        ----------
        parameter : (p,) ndarray
            Parameter value :math:`\bfmu`.
        state : (r,) ndarray
            State vector :math:`\qhat`.
        input_ : (m,) ndarray or None
            Input vector :math:`\u`.

        Returns
        -------
        jac : (r, r) ndarray
            State Jacobian of the right-hand side of the model.

        Notes
        -----
        For repeated ``jacobian()`` calls with the same parameter value, use
        :meth:`evaluate` to first get the nonparametric model corresponding
        to the parameter value.

        .. code-block::

           # Instead of this...
           >>> jacs = [parametric_model.jacobian(parameter, q, input_)
           ...         for q in list_of_states]
           # ...it is faster to do this.
           >>> model_at_parameter = parametric_model.evaluate(parameter)
           >>> jacs = [model_at_parameter.jacobian(q, input_)
           ...         for q in list_of_states]
        """
        return super().jacobian(parameter, state, input_)

    def predict(self, parameter, state0, niters, inputs=None):
        r"""Step forward the discrete dynamical system
        ``niters`` steps. Essentially, this amounts to the following.

        .. code-block:: python

           >>> states[:, 0] = state0
           >>> states[:, 1] = model.rhs(parameter, states[:, 0], inputs[:, 0])
           >>> states[:, 2] = model.rhs(parameter, states[:, 1], inputs[:, 1])
           ...                                     # Repeat `niters` times.

        Parameters
        ----------
        parameter : (p,) ndarray
            Parameter value :math:`\bfmu`.
        state0 : (r,) ndarray
            Initial state.
        niters : int
            Number of times to step the system forward.
        inputs : (m, niters-1) ndarray or None
            Inputs for the next ``niters - 1`` time steps.

        Returns
        -------
        states : (r, niters) ndarray
            Solution to the system, including the initial condition ``state0``.

        Notes
        -----
        For repeated ``predict()`` calls with the same parameter value, use
        :meth:`evaluate` to first get the nonparametric model corresponding
        to the parameter value.
        """
        return super().predict(parameter, state0, niters, inputs=inputs)


class _ParametricContinuousMixin:
    """Mixin class for parametric models of system of ODEs."""

    _ModelClass = _FrozenContinuousModel

    def fit(self, parameters, states, ddts, inputs=None):
        r"""Learn the model operators from data.

        The operators are inferred by solving the regression problem

        .. math::
           \min_{\Ohat^{(1)},\ldots,\Ohat^{(s)}}
           \sum_{i=1}^{s}\sum_{j=0}^{k_{i}-1}\left\|
           \Ophat(\qhat_{i,j}, \u_{i,j}; \bfmu_i) - \dot{\qhat}_{i,j}
           \right\|_2^2
           = \min_{\Ohat^{(1)},\ldots,\Ohat^{(s)}}
           \sum_{i=1}^{s}\left\|
           \D^{(i)}(\Ohat^{(i)})\trp
           - [~\dot{\qhat}_{i,0}~~\cdots~~\dot{\qhat}_{i,k_{i}-1}~]\trp
           \right\|_F^2

        where
        :math:`\ddt\qhat(t; \bfmu) = \Ophat(\qhat(t; \bfmu), \u(t); \bfmu)
        = \Ohat(\bfmu)\d(\qhat(t; \bfmu), \u(t))` is the model and

        * :math:`\qhat_{i,j}\in\RR^r` is the :math:`j`-th measurement of the
          state corresponding to training parameter value :math:`\bfmu_i`,
        * :math:`\u_{i,j}\in\RR^m` is the :math:`j`-th measurement of the
          input corresponding to training parameter value :math:`\bfmu_i`,
        * :math:`\dot{qhat}_{i,j}\in\RR^r` is a measurement of the time
          derivative of the state :math:`\qhat_{i,j}`, i.e.,
          :math:`\dot{\qhat}_{i,j} = \ddt\qhat(t; \bfmu_i)\big|_{t=t_{i,j}}`
          where `\qhat_{i,j} = \qhat(t_{i,j};\bfmu_i)`,
        * :math:`\Ohat^{(i)} = \Ohat(\bfmu_i)` is the operator matrix
          evaluated at training parameter value :math:`\bfmu_i`, and
        * :math:`\D^{(i)}` is the data matrix for data corresponding to
          training parameter value :math:`\bfmu_i`, given by
          :math:`[~\d(\qhat_{i,0},\u_{i,0})~~\cdots~~
          \d(\qhat_{i,k_i-1},\u_{i,k_i-1})~]\trp`.

        Because all operators in this model are interpolatory, the
        least-squares problem decouples into :math:`s` individual regressions.
        That is, for :math:`i = 1, \ldots, s`, we solve (independently) the
        regressions

        .. math::
           \min_{\Ohat^{(i)}}\left\|
           \D^{(i)}(\Ohat^{(i)})\trp
           - [~\dot{\qhat}_{i,0}~~\cdots~~\dot{\qhat}_{i,k_{i}-1}~]\trp
           \right\|_F^2

        and define the full operator matrix via elementwise interpolation,

        .. math::
           \Ohat(\bfmu) = \textrm{interpolate}(
           (\bfmu_1, \Ohat^{(i)}), \ldots, (\bfmu_s, \Ohat^{(s)}); \bfmu).

        The strategy for solving the regression, as well as any additional
        regularization or constraints, are specified by the ``solver`` in
        the constructor.

        Parameters
        ----------
        parameters : list of s scalars or (p,) 1D ndarrays
            Parameter values for which the operator entries are known
            or will be inferred from data.
        states : list of s (r, k_i) ndarrays
            Snapshot training data. Each array ``states[i]`` is the data
            corresponding to parameter value ``parameters[i]``; each column
            ``states[i][:, j]`` is a single snapshot.
        ddts : list of s (r, k_i) ndarrays
            Snapshot time derivative data. Each array ``ddts[i]`` is the data
            corresponding to parameter value ``parameters[i]``; each column
            ``ddts[i][:, j]`` corresponds to the snapshot ``states[i][:, j]``.
        inputs : list of s (m, k_i) or (k_i,) ndarrays, or None
            Input training data. Each array ``inputs[i]`` is the data
            corresponding to parameter value ``parameters[i]``; each column
            ``inputs[i][:, j]`` corresponds to the snapshot
            ``states[i][:, j]``.
            May be a two-dimensional array if :math:`m=1` (scalar input).
            Only required if one or more model operators depend on inputs.

        Returns
        -------
        self
        """
        return super().fit(parameters, states, ddts, inputs=inputs)

    def rhs(self, t, parameter, state, input_func=None):
        r"""Evaluate the right-hand side of the model by applying each operator
        and summing the results.

        This is the right-hand side of the model, i.e., the function
        :math:`\fhat(\qhat(t), \u(t); \bfmu)` where the model is given by
        :math:`\ddt \qhat(t; \bfmu) = \fhat(\qhat(t), \u(t); \bfmu)`.

        Parameters
        ----------
        t : float
            Time :math:`t`, a scalar.
        parameter : (p,) ndarray
            Parameter value :math:`\bfmu`.
        state : (r,) ndarray
            State vector :math:`\qhat(t)` corresponding to time ``t``.
        input_func : callable(float) -> (m,), or None
            Input function that maps time ``t`` to the input vector
            :math:`\u(t)`.

        Returns
        -------
        dqdt : (r,) ndarray
            Evaluation of the right-hand side of the model.

        Notes
        -----
        For repeated ``rhs()`` calls with the same parameter value, use
        :meth:`evaluate` to first get the nonparametric model corresponding
        to the parameter value.

        .. code-block::

           # Instead of this...
           >>> values = [parametric_model.rhs(t, parameter, q, input_func)
           ...           for t, q in zip(times, states)]
           # ...it is faster to do this.
           >>> model_at_parameter = parametric_model.evaluate(parameter)
           >>> values = [model_at_parameter.rhs(t, parameter, q, input_func)
           ...           for t, q in zip(times, states)]
        """
        return super().rhs(parameter, t, state, input_func)

    def jacobian(self, t, parameter, state, input_func=None):
        r"""Sum the state Jacobian of each model operator.

        This the derivative of the right-hand side of the model with respect
        to the state, i.e., the function
        :math:`\ddqhat\fhat(\qhat(t), \u(t); \bfmu)` where the model is given
        by :math:`\ddt\qhat(t; \bfmu) = \fhat(\qhat(t), \u(t); \bfmu)`.

        Parameters
        ----------
        t : float
            Time :math:`t`, a scalar.
        parameter : (p,) ndarray
            Parameter value :math:`\bfmu`.
        state : (r,) ndarray
            State vector :math:`\qhat(t)` corresponding to time ``t``.
        input_func : callable(float) -> (m,), or None
            Input function that maps time ``t`` to the input vector
            :math:`\u(t)`.

        Returns
        -------
        jac : (r, r) ndarray
            State Jacobian of the right-hand side of the model.

        Notes
        -----
        For repeated ``jacobian()`` calls with the same parameter value, use
        :meth:`evaluate` to first get the nonparametric model corresponding
        to the parameter value.

        .. code-block::

           # Instead of this...
           >>> jacs = [parametric_model.jacobian(t, parameter, q, input_func)
           ...         for t, q in zip(times, states)]
           # ...it is faster to do this.
           >>> model_at_parameter = parametric_model.evaluate(parameter)
           >>> jacs = [model_at_parameter.jacobian(t, parameter, q, input_func)
           ...         for t, q in zip(times, states)]
        """
        return super().jacobian(parameter, t, state, input_func)

    def predict(self, parameter, state0, t, input_func=None, **options):
        r"""Solve the system of ordinary differential equations.
        This method wraps :func:`scipy.integrate.solve_ivp()`.

        Parameters
        ----------
        parameter : (p,) ndarray
            Parameter value :math:`\bfmu`.
        state0 : (r,) ndarray
            Initial state vector.
        t : (nt,) ndarray
            Time domain over which to integrate the model.
        input_func : callable or (m, nt) ndarray
            Input as a function of time (preferred) or the input values at the
            times ``t``. If given as an array, cubic spline interpolation on
            the known data points is used as needed.
        options
            Arguments for :func:`scipy.integrate.solve_ivp()`.
            Common options:

            * **method : str** ODE solver for the model.

              * ``'RK45'`` (default): Explicit Runge--Kutta method
                of order 5(4).
              * ``'RK23'``: Explicit Runge--Kutta method
                of order 3(2).
              * ``'Radau'``: Implicit Runge--Kutta method
                of the Radau IIA family of order 5.
              * ``'BDF'``: Implicit multi-step variable-order (1 to 5) method
                based on a backward differentiation formula for the derivative.
              * ``'LSODA'``: Adams/BDF method with automatic stiffness
                detection and switching.

            * **max_step : float** Maximimum allowed integration step size.

        Returns
        -------
        states : (r, nt) ndarray
            Computed solution to the system over the time domain ``t``.

        Notes
        -----
        For repeated ``predict()`` calls with the same parameter value, use
        :meth:`evaluate` to first get the nonparametric model corresponding
        to the parameter value.
        """
        return super().predict(
            parameter, state0, t, input_func=input_func, **options
        )


# Public classes ==============================================================
class ParametricDiscreteModel(_ParametricDiscreteMixin, _ParametricModel):
    r"""Parametric discrete dynamical system model
    :math:`\qhat(\bfmu)_{j+1} = \Ophat(\qhat(\bfmu)_{j}, \u_{j}; \bfmu)`.

    Here,

    * :math:`\qhat(\bfmu)_j\in\RR^{r}` is the :math:`j`-th iteration
      of the model state,
    * :math:`\u_j\in\RR^{m}` is the (optional) corresponding input, and
    * :math:`\bfmu\in\RR^{p}\in\RR^{p}` is the parameter vector.


    The structure of :math:`\Ophat` is specified through the
    ``operators`` attribute.

    Parameters
    ----------
    operators : list of :mod:`opinf.operators` objects
        Operators comprising the terms of the model.
    """

    pass


class ParametricContinuousModel(_ParametricContinuousMixin, _ParametricModel):
    r"""Parametric system of ordinary differential equations
    :math:`\ddt\qhat(t; \bfmu) = \fhat(\qhat(t; \bfmu), \u(t); \bfmu)`.

    Here,

    * :math:`\qhat(t;\bfmu)\in\RR^{r}` is the model state,
    * :math:`\u(t)\in\RR^{m}` is the (optional) input, and
    * :math:`\bfmu\in\RR^{p}\in\RR^{p}` is the parameter vector.

    The structure of :math:`\fhat` is specified through the
    ``operators`` argument.

    Parameters
    ----------
    operators : list of :mod:`opinf.operators` objects
        Operators comprising the terms of the model.
    """

    pass


# Special case: completely interpolation-based models =========================
class _InterpModel(_ParametricModel):
    """Base class for parametric monolithic models where all operators MUST be
    interpolation-based parametric operators. In this special case, the
    inference problems completely decouple by training parameter.

    Parameters
    ----------
    operators : list of :mod:`opinf.operators` objects
        Operators comprising the terms of the model.
    solver : :mod:`opinf.lstsq` object or float > 0 or None
        Solver for the least-squares regression. Defaults:

        * ``None``: :class:`opinf.lstsq.PlainSolver`, SVD-based solve
            without regularization.
        * float > 0: :class:`opinf.lstsq.L2Solver`, SVD-based solve with
            scalar Tikhonov regularization.
    InterpolatorClass : type or None
        Class for the elementwise interpolation. Must obey the syntax

        .. code-block:: python

           >>> interpolator = InterpolatorClass(data_points, data_values)
           >>> interpolator_evaluation = interpolator(new_data_point)

        This can be, e.g., a class from :mod:`scipy.interpolate`.
        If ``None`` (default), use :class:`scipy.interpolate.CubicSpline`
        for one-dimensional parameters and
        :class:`scipy.interpolate.LinearNDInterpolator` otherwise.
    """

    @property
    def _ModelFitClass(self):
        """Parent of ModelClass that has a callable ``fit()`` method."""
        return self._ModelClass.__bases__[-1]

    def __init__(self, operators, solver=None, InterpolatorClass=None):
        """Define the model structure and set the interpolator class."""
        _ParametricModel.__init__(self, operators, solver=solver)
        self.set_interpolator(InterpolatorClass)
        self._submodels = None
        self._training_parameters = None

    @classmethod
    def _from_models(cls, parameters, models, InterpolatorClass: type = None):
        """Interpolate a collection of non-parametric models.

        Parameters
        ----------
        parameters : list of s scalars or (p,) 1D ndarrays
            Parameter values for which the operator entries are known.
        models : list of s :mod:`opinf.models` objects
            Nonparametric models with fully populated operator entries.
        InterpolatorClass : type or None
            Class for the elementwise interpolation. Must obey the syntax

               >>> interpolator = InterpolatorClass(data_points, data_values)
               >>> interpolator_evaluation = interpolator(new_data_point)

            This can be, e.g., a class from :mod:`scipy.interpolate`.
            If ``None`` (default), use :class:`scipy.interpolate.CubicSpline`
            for one-dimensional parameters and
            :class:`scipy.interpolate.LinearNDInterpolator` otherwise.
        """
        # Check for consistency in the model operators.
        opclasses = [type(op) for op in models[0].operators]
        for mdl in models:
            # Operator count and type.
            if len(mdl.operators) != len(opclasses):
                raise ValueError(
                    "models not aligned (inconsistent number of operators)"
                )
            for ell, op in enumerate(mdl.operators):
                if not isinstance(op, opclasses[ell]):
                    raise ValueError(
                        "models not aligned (inconsistent operator types)"
                    )
            # Entries are set.
            mdl._check_is_trained()

        # Extract the operators from the individual models.
        return cls(
            operators=[
                oputils.nonparametric_to_interpolated(OpClass)._from_operators(
                    training_parameters=parameters,
                    operators=[mdl.operators[ell] for mdl in models],
                    InterpolatorClass=InterpolatorClass,
                )
                for ell, OpClass in enumerate(opclasses)
            ],
            InterpolatorClass=InterpolatorClass,
        )

    def set_interpolator(self, InterpolatorClass):
        """Set the interpolator for the operator entries.

        Parameters
        ----------
        InterpolatorClass : type
            Class for the elementwise interpolation. Must obey the syntax

                >>> interpolator = InterpolatorClass(data_points, data_values)
                >>> interpolator_evaluation = interpolator(new_data_point)

            This can be, e.g., a class from :mod:`scipy.interpolate`.
            If ``None`` (default), use :class:`scipy.interpolate.CubicSpline`
            for one-dimensional parameters and
            :class:`scipy.interpolate.LinearNDInterpolator` otherwise.
        """
        if InterpolatorClass is not None:
            for op in self.operators:
                op.set_interpolator(InterpolatorClass)
        self.__InterpolatorClass = InterpolatorClass

    # Properties: operators ---------------------------------------------------
    _operator_abbreviations = {
        "c": _operators.InterpConstantOperator,
        "A": _operators.InterpLinearOperator,
        "H": _operators.InterpQuadraticOperator,
        "G": _operators.InterpCubicOperator,
        "B": _operators.InterpInputOperator,
        "N": _operators.InterpStateInputOperator,
    }

    def _isvalidoperator(self, op):
        """Only interpolated parametric operators are allowed."""
        return oputils.is_interpolated(op)

    # Fitting -----------------------------------------------------------------
    def _assemble_data_matrix(self, *args, **kwargs):  # pragma: no cover
        """Assemble the data matrix for operator inference."""
        raise NotImplementedError(
            "_assemble_data_matrix() not used by this class"
        )

    def _extract_operators(self, *args, **kwargs):  # pragma: no cover
        """Unpack the operator matrix and populate operator entries."""
        raise NotImplementedError(
            "_extract_operators() not used by this class"
        )

    def _fit_solver(self, parameters, states, lhs, inputs=None):
        """Construct a solver for the operator inference least-squares
        regression.
        """
        (
            parameters_,
            states_,
            lhs_,
            inputs_,
        ) = self._process_fit_arguments(parameters, states, lhs, inputs)
        n_datasets = len(parameters)

        # Distribute training data to individual OpInf problems.
        nonparametric_models = []
        for i in range(n_datasets):
            model_i = self._ModelFitClass(
                operators=[
                    op._OperatorClass(
                        op.entries[i] if op.entries is not None else None
                    )
                    for op in self.operators
                ],
                solver=self.solver.copy(),
            )
            model_i._fit_solver(
                states_[i],
                lhs_[i],
                inputs_[i],
            )
            nonparametric_models.append(model_i)

        self.solvers = [mdl.solver for mdl in nonparametric_models]
        self._submodels = nonparametric_models
        self._training_parameters = parameters_

    def refit(self):
        """Evaluate the least-squares solver and process the results."""
        if self._submodels is None:
            raise RuntimeError("model solvers not set, call fit() first")

        # Solve each independent subproblem.
        # TODO: parallelize?
        for submodel in self._submodels:
            submodel.refit()

        # Interpolate the resulting operators.
        for ell, op in enumerate(self.operators):
            op._clear()
            op.set_training_parameters(self._training_parameters)
            op.set_entries(
                [mdl.operators[ell].entries for mdl in self._submodels]
            )

        return self

    # Model persistence -------------------------------------------------------
    def save(self, savefile: str, overwrite: bool = False) -> None:
        """Serialize the model, saving it in HDF5 format.
        The model can be recovered with the :meth:`load()` class method.

        Parameters
        ----------
        savefile : str
            File to save to, with extension ``.h5`` (HDF5).
        overwrite : bool
            If ``True`` and the specified ``savefile`` already exists,
            overwrite the file. If ``False`` (default) and the specified
            ``savefile`` already exists, raise an error.
        """
        with utils.hdf5_savehandle(savefile, overwrite=overwrite) as hf:
            # Metadata.
            meta = hf.create_dataset("meta", shape=(0,))
            meta.attrs["num_operators"] = len(self.operators)
            meta.attrs["r"] = (
                int(self.state_dimension) if self.state_dimension else 0
            )
            meta.attrs["m"] = (
                int(self.input_dimension) if self.input_dimension else 0
            )

            # Interpolator class.
            suppress_warnings = False
            InterpolatorClassName = (
                "NoneType"
                if self.__InterpolatorClass is None
                else self.__InterpolatorClass.__name__
            )
            meta.attrs["InterpolatorClass"] = InterpolatorClassName
            if InterpolatorClassName != "NoneType" and not hasattr(
                spinterpolate, InterpolatorClassName
            ):
                warnings.warn(
                    "cannot serialize InterpolatorClass "
                    f"'{InterpolatorClassName}', must pass in the class "
                    "when calling load()",
                    errors.OpInfWarning,
                )
                suppress_warnings = True

            # Operator data.
            hf.create_dataset(
                "indices_infer", data=self._indices_of_operators_to_infer
            )
            hf.create_dataset(
                "indices_known", data=self._indices_of_known_operators
            )
            with warnings.catch_warnings():
                if suppress_warnings:
                    warnings.simplefilter("ignore", errors.OpInfWarning)
                for i, op in enumerate(self.operators):
                    op.save(hf.create_group(f"operator_{i}"))

    @classmethod
    def load(cls, loadfile: str, InterpolatorClass: type = None):
        """Load a serialized model from an HDF5 file, created previously from
        the :meth:`save()` method.

        Parameters
        ----------
        loadfile : str
            Path to the file where the operator was stored via :meth:`save()`.
        InterpolatorClass : type
            Class for the elementwise interpolation. Must obey the syntax

               >>> interpolator = InterpolatorClass(data_points, data_values)
               >>> interpolator_evaluation = interpolator(new_data_point)

            Not required if the saved operator utilizes a class from
            :mod:`scipy.interpolate`.

        Returns
        -------
        op : _Operator
            Initialized operator object.
        """
        with utils.hdf5_loadhandle(loadfile) as hf:
            # Load metadata.
            num_operators = int(hf["meta"].attrs["num_operators"])
            indices_infer = [int(i) for i in hf["indices_infer"][:]]
            indices_known = [int(i) for i in hf["indices_known"][:]]

            # Get the InterpolatorClass.
            SavedClassName = hf["meta"].attrs["InterpolatorClass"]
            if InterpolatorClass is None:
                # Load from scipy.interpolate.
                if hasattr(spinterpolate, SavedClassName):
                    InterpolatorClass = getattr(spinterpolate, SavedClassName)
                elif SavedClassName != "NoneType":
                    raise ValueError(
                        f"unknown InterpolatorClass '{SavedClassName}', "
                        f"call load({loadfile}, {SavedClassName})"
                    )
            else:
                # Warn the user if the InterpolatorClass does not match.
                if SavedClassName != (
                    InterpolatorClassName := InterpolatorClass.__name__
                ):
                    warnings.warn(
                        f"InterpolatorClass={InterpolatorClassName} does not "
                        f"match loadfile InterpolatorClass '{SavedClassName}'",
                        errors.OpInfWarning,
                    )

            # Load operators.
            ops = []
            for i in range(num_operators):
                gp = hf[f"operator_{i}"]
                OpClassName = gp["meta"].attrs["class"]
                OpClass = getattr(_operators, OpClassName)
                ops.append(OpClass.load(gp, InterpolatorClass))

            # Construct the model.
            model = cls(ops)
            model._indices_of_operators_to_infer = indices_infer
            model._indices_of_known_operators = indices_known
            if r := int(hf["meta"].attrs["r"]):
                model.state_dimension = r
            if (m := int(hf["meta"].attrs["m"])) and model._has_inputs:
                model.input_dimension = m

        return model

    def copy(self):
        """Make a copy of the model."""
        return self.__class__(
            operators=[op.copy() for op in self.operators],
            InterpolatorClass=self.__InterpolatorClass,
        )


class InterpDiscreteModel(_ParametricDiscreteMixin, _InterpModel):
    r"""Parametric discrete dynamical system model
    :math:`\qhat(\bfmu)_{j+1} = \fhat(\qhat(\bfmu)_{j}, \u_{j}; \bfmu)`
    where the parametric dependence is handled by elementwise interpolation.

    Here,

    * :math:`\qhat_j\in\RR^{r}` is the :math:`j`-th iteration
      of the model state,
    * :math:`\u_j\in\RR^{m}` is the (optional) corresponding input, and
    * :math:`\bfmu\in\RR^{p}\in\RR^{p}` is the parameter vector.


    The structure of :math:`\fhat` is specified through the
    ``operators`` attribute.

    Parameters
    ----------
    operators : list of :mod:`opinf.operators` objects
        Operators comprising the terms of the model.
        For this class, these must be interpolated parametric operators.
    InterpolatorClass : type or None
        Class for the elementwise interpolation. Must obey the syntax

            >>> interpolator = InterpolatorClass(data_points, data_values)
            >>> interpolator_evaluation = interpolator(new_data_point)

        This can be, e.g., a class from :mod:`scipy.interpolate`.
        If ``None`` (default), use :class:`scipy.interpolate.CubicSpline`
        for one-dimensional parameters and
        :class:`scipy.interpolate.LinearNDInterpolator` otherwise.
    """

    pass


class InterpContinuousModel(
    _ParametricContinuousMixin,
    _InterpModel,
):
    r"""Parametric system of ordinary differential equations
    :math:`\ddt\qhat(t; \bfmu) = \fhat(\qhat(t; \bfmu), \u(t); \bfmu)` where
    the parametric dependence is handled by elementwise interpolation.

    Here,

    * :math:`\qhat(t;\bfmu)\in\RR^{r}` is the model state,
    * :math:`\u(t)\in\RR^{m}` is the (optional) input, and
    * :math:`\bfmu\in\RR^{p}\in\RR^{p}` is the parameter vector.

    The structure of :math:`\fhat` is specified through the
    ``operators`` argument.

    Parameters
    ----------
    operators : list of :mod:`opinf.operators` objects
        Operators comprising the terms of the model.
        For this class, these must be interpolated parametric operators.
    InterpolatorClass : type or None
        Class for the elementwise interpolation. Must obey the syntax

            >>> interpolator = InterpolatorClass(data_points, data_values)
            >>> interpolator_evaluation = interpolator(new_data_point)

        This can be, e.g., a class from :mod:`scipy.interpolate`.
        If ``None`` (default), use :class:`scipy.interpolate.CubicSpline`
        for one-dimensional parameters and
        :class:`scipy.interpolate.LinearNDInterpolator` otherwise.
    """

    pass


# Deprecations ================================================================
class InterpolatedDiscreteModel(InterpDiscreteModel):
    def __init__(self, operators, solver=None, InterpolatorClass=None):
        warnings.warn(
            "InterpolatedDiscreteModel has been renamed "
            "and will be removed in an upcoming release, use "
            "InterpDiscreteModel",
            DeprecationWarning,
        )
        InterpDiscreteModel.__init__(
            self,
            operators=operators,
            solver=solver,
            InterpolatorClass=InterpolatorClass,
        )


class InterpolatedContinuousModel(InterpContinuousModel):
    def __init__(self, operators, solver=None, InterpolatorClass=None):
        warnings.warn(
            "InterpolatedContinuousModel has been renamed "
            "and will be removed in an upcoming release, use "
            "InterpContinuousModel",
            DeprecationWarning,
        )
        InterpContinuousModel.__init__(
            self,
            operators=operators,
            solver=solver,
            InterpolatorClass=InterpolatorClass,
        )
